#coverage:ignore
""" Determine costs for SF decomposition in QC """
from typing import Tuple
import numpy as np
from numpy.lib.scimath import arccos, arcsin  # has analytic continutn to cplx
from openfermion.resource_estimates.utils import QR, QI, QR2, power_two


def compute_cost(n: int,
                 lam: float,
                 dE: float,
                 L: int,
                 chi: int,
                 stps: int,
                 verbose: bool = False) -> Tuple[int, int, int]:
    """ Determine fault-tolerant costs using SF decomposition in quantum chem

    Args:
        n (int) - the number of spin-orbitals
        lam (float) - the lambda-value for the Hamiltonian
        dE (float) - allowable error in phase estimation
        L (int) - the rank of the first decomposition
        chi (int) - equivalent to aleph_1 and aleph_2 in the document, the
            number of bits for the representation of the coefficients
        stps (int) - an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superposn state
        verbose (bool) - do additional printing of intermediates?

    Returns:
        step_cost (int) - Toffolis per step
        total_cost (int) - Total number of Toffolis
        total_qubit_count (int) - Total qubit count
    """

    # Number of trailing zeros
    eta = power_two(L + 1)

    # Number of qubits for the first register
    nL = np.ceil(np.log2(L + 1))

    # Number of qubits for p and q registers
    nN = np.ceil(np.log2(n // 2))

    oh = [0] * 20
    for p in range(20):
        # JJG note: arccos arg may be > 1
        v = np.round(np.power(2,p+1) / (2 * np.pi) * arccos(np.power(2,nL) /\
            np.sqrt((L + 1)/2**eta)/2))
        oh[p] = np.real(stps * (1 / (np.sin(3 * arcsin(np.cos(v * 2 * np.pi / \
            np.power(2,p+1)) * \
            np.sqrt((L + 1)/2**eta) / np.power(2,nL)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br = int(np.argmin(oh) + 1)

    # The following costs are given in the list of steps on pages 43 and 44.

    # Cost of preparing the equal superposition state on the 1st register in 1a
    cost1a = 2 * (3 * nL - 3 * eta + 2 * br - 9)

    # We have added two qubits
    bL = nL + chi + 2

    # QROM costs for first register preparation in step 1(b)
    cost1b = QR(L + 1, bL)[-1] + QI(L + 1)[-1]

    # The inequality test of cost chi in step 1(c) and the controlled swap with
    # cost nL + 1 in step 1(d) (and their inversions).
    cost1cd = 2 * (chi + nL + 1)

    oh = [0] * 20
    nprime = int(n**2 // 8 + n // 4)
    for p in range(20):
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, 2 * nN) / np.sqrt(nprime) / 2))
        oh[p] = np.real(20000 * (1 / (np.sin(3 * arcsin(np.cos(v * 2 * np.pi / \
            np.power(2, p+1)) * \
            np.sqrt(nprime) / np.power(2,2*nN)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation for preparing the next equal
    # superposition states
    br = int(np.argmin(oh) + 1)

    # Cost of preparing equal superposition of p and q registers in step 2(a).
    cost2a = 4 * (6 * nN + 2 * br - 7)

    # Cost of computing contiguous register in step 2 (b).
    cost2b = 4 * (nN**2 + nN - 1)

    # Number of coefficients for first state preparation on p & q.
    n1 = (L + 1) * nprime

    # Number of coefficients for second state preparation on p & q.
    n2 = L * nprime

    # The output size for the QROM for the second state preparation.
    bp = int(2 * nN + chi + 2)

    # Cost of QROMs for state preparation on p and q in step 2 (c).
    cost2c = QR2(L + 1, nprime, bp)[-1] + QI(n1)[-1] + QR2(L, nprime,
                                                           bp)[-1] + QI(n2)[-1]

    # The cost of the inequality test and controlled swap for the quantum alias
    # sampling in steps 2 (d) and (e).
    cost2de = 4 * (chi + 2 * nN)

    # Swapping the p & q registers for symmetry in step 3.  This needs to be
    # done and inverted twice, hence the factor of 4.
    cost3 = 4 * nN

    # The SELECT operation in step 4, which needs to be done twice.
    cost4 = 4 * n - 8

    # The reflection used on the 2nd register in the middle, with cost in step 6
    cost6 = 2 * nN + chi + 3

    # Step 7 involves performing steps 2 to 5 again, which are accounted for
    # above, but there is one extra Toffoli for checking that l <> 0.
    cost7 = 1

    # The total reflection for the quantum walk in step 9.
    cost9 = nL + 2 * nN + 2 * chi + 2

    # The two Toffolis for the control for phase estimation and iteration,
    # given in step 10.
    cost10 = 2

    # The number of steps.
    iters = np.ceil(np.pi * lam / (dE * 2))

    # The total Toffoli costs for a step.
    cost = cost1a + cost1b + cost1cd + cost2a + cost2b + cost2c + cost2de + \
           cost3 + cost4 + cost6 + cost7 + cost9 + cost10

    # Control for phase estimation and its iteration
    ac1 = 2 * np.ceil(np.log2(iters)) - 1

    # System qubits
    ac2 = n

    # First ell register that rotated and the flag for success
    ac3 = nL + 2

    # State preparation on the first register
    ac4 = nL + 2 * chi + 3

    # For p and q registers, the rotated qubit and the success flag
    ac5 = 2 * nN + 2

    # The size of the contiguous register
    ac6 = np.ceil(np.log2(nprime))

    kp = QR2(L + 1, nprime, bp)[:2]

    # The equal superposition state for the second state preparation
    ac7 = chi

    # The phase gradient state
    ac8 = br

    # The QROM on the p & q registers
    ac9 = kp[0] * kp[1] * bp + np.ceil(np.log2(
        (L + 1) / kp[0])) + np.ceil(np.log2(nprime / kp[1]))

    if verbose:
        print("[*] Top of routine")
        print("  [+] eta = ", eta)
        print("  [+] nL = ", nL)
        print("  [+] nN = ", nN)

    total_qubit_count = ac1 + ac2 + ac3 + ac4 + ac5 + ac6 + ac7 + ac8 + ac9

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert total_qubit_count.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    total_qubit_count = int(total_qubit_count)

    return step_cost, total_cost, total_qubit_count



#coverage:ignore
""" Compute lambda for single low rank factorization method of Berry, et al """

import numpy as np
from openfermion.resource_estimates.molecule import pyscf_to_cas


def compute_lambda(pyscf_mf, sf_factors):
    """ Compute lambda for Hamiltonian using SF method of Berry, et al.

    Args:
        pyscf_mf - PySCF mean field object
        sf_factors (ndarray) - (N x N x rank) array of SF factors from rank
            reduction of ERI

    Returns:
        lambda_tot (float) - lambda value for the single factorized Hamiltonian
    """

    # grab tensors from pyscf_mf object
    h1, eri_full, _, _, _ = pyscf_to_cas(pyscf_mf)

    # Effective one electron operator contribution
    T = h1 - 0.5 * np.einsum("pqqs->ps", eri_full, optimize=True) +\
        np.einsum("pqrr->pq", eri_full, optimize = True)

    lambda_T = np.sum(np.abs(T))

    # Two electron operator contributions
    lambda_W = 0.25 * np.einsum(
        "ijP,klP->", np.abs(sf_factors), np.abs(sf_factors), optimize=True)
    lambda_tot = lambda_T + lambda_W

    return lambda_tot



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# coverage:ignore
""" Determine costs for THC decomposition in QC """
from typing import Tuple, Union
import numpy as np
from numpy.lib.scimath import arccos, arcsin
from sympy import factorint
from openfermion.resource_estimates.utils import QI

from openfermion.resource_estimates.pbc.resources.data_types import (
    ResourceEstimates,)
from openfermion.resource_estimates.pbc.resources.qrom import QR3


def compute_cost(
        num_spin_orbs: int,
        lambda_tot: float,
        thc_dim: int,
        kmesh: list[int],
        dE_for_qpe: float = 0.0016,
        chi: int = 10,
        beta: Union[int, None] = None,
) -> ResourceEstimates:
    """Determine fault-tolerant costs using THC factorization representation of
        symmetry adapted integrals.

    Light wrapper around _compute_cost.

    Arguments:
        num_spin_orbs: the number of spin-orbitals
        lambda_tot: the lambda-value for the Hamiltonian
        kmesh: kpoint mesh.
        thc_dim: THC dimension (M).
        dE_for_qpe: allowable error in phase estimation
        chi: the number of bits for the representation of the coefficients
        beta: the number of bits for controlled rotations
    Returns:
        resources: THC factorized resource estimates
    """
    # run once to determine stps parameter
    thc_costs = _compute_cost(
        n=num_spin_orbs,
        lam=lambda_tot,
        dE=dE_for_qpe,
        chi=chi,
        beta=beta,
        M=thc_dim,
        Nkx=kmesh[0],
        Nky=kmesh[1],
        Nkz=kmesh[2],
        stps=20_000,  # not used
    )
    resources = ResourceEstimates(
        toffolis_per_step=thc_costs[0],
        total_toffolis=thc_costs[1],
        logical_qubits=thc_costs[2],
    )
    return resources


def _compute_cost(
        n: int,
        lam: float,
        dE: float,
        chi: int,
        beta: int,
        M: int,
        Nkx: int,
        Nky: int,
        Nkz: int,
        stps: int,
        verbose: bool = False,
) -> Tuple[int, int, int]:
    """Determine fault-tolerant costs using THC decomposition in quantum chem

    Arguments:
        n: the number of spin-orbitals
        lam: the lambda-value for the Hamiltonian
        dE: allowable error in phase estimation
        chi: equivalent to aleph in the document, the number of bits for
            the representation of the coefficients
        beta: equivalent to beth in the document, the number of bits for
            the rotations
        M: the dimension for the THC decomposition
        Nkx: This is the number of values of k for the k - point sampling,
                with each component
        Nky: This is the number of values of k for the k - point sampling,
                with each component
        Nkz: This is the number of values of k for the k - point sampling,
                with each component
        stps: an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superpositn state

    Returns:
        step_cost: Toffolis per step
        total_cost: Total number of Toffolis
        ancilla_cost: Total ancilla cost
    """
    nk = (max(np.ceil(np.log2(Nkx)), 1) + max(np.ceil(np.log2(Nky)), 1) +
          max(np.ceil(np.log2(Nkz)), 1))
    Nk = Nkx * Nky * Nkz

    # (*Temporarily set as number of even numbers.*)
    nc = 3 - Nkx % 2 - Nky % 2 - Nkz % 2

    # The number of steps needed
    iters = np.ceil(np.pi * lam / (2 * dE))

    # This is the number of distinct items of data we need to output
    # see Eq. (28).*)
    d = int(32 * (Nk + 2**nc) * M**2 + n * Nk / 2)

    # The number of bits used for the contiguous register
    nc = np.ceil(np.log2(d))

    nM = np.ceil(np.log2(M))

    # The output size is 2* Log[M] for the alt values, chi for the keep value,
    # and 2 for the two sign bits.
    m = 2 * (2 * nM + nk + 8) + chi

    oh = [0] * 20
    for p in range(20):
        # arccos arg may be > 1
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nc) / np.sqrt(d) / 2))
        oh[p] = stps * (1 / (np.sin(3 * arcsin(
            np.cos(v * 2 * np.pi / np.power(2, p + 1)) * np.sqrt(d) /
            np.power(2, nc)))**2) - 1) + 4 * (p + 1)

    # Set it to be the number of bits that minimises the cost, usually 7.
    # Python is 0-index, so need to add the one back in vs mathematica nb
    br = np.argmin(oh) + 1

    if d % 2 == 0:
        factors = factorint(int(d))
        eta = factors[min(list(sorted(factors.keys())))]
    else:
        eta = 0

    cp1 = 2 * (3 * nc - 3 * eta + 2 * br - 9)

    # This is the cost of the QROM for the state preparation in step 3 and its
    cp3 = QR3(d, m)[1] + QI(d)[1]

    # The cost for the inequality test in step 4 and its inverse.
    cp4 = 2 * chi

    # the cost of inequality test and controlled swap of mu and nu registers
    cp5 = 2 * (2 * nM + nk + 7)

    # The cost of an inequality test and controlled - swap of the mu and nu
    # registers
    cp6 = 4 * nM + 12

    CPCP = cp1 + cp3 + cp4 + cp5 + cp6

    # The cost of preparing the k superposition. The 7 here is the assumed
    # number # of bits for the ancilla rotation which makes the probability of
    # failure negligible.
    cks = 4 * (Nkx + Nky + Nkz + 8 * nk + 6 * 7 - 24)

    # The cost of the arithmetic computing k - Q, and controlling swaps for
    # the one-body term
    cka = 12 * nk

    # This is the cost of swapping based on the spin register
    cs1 = 3 * n * Nk / 2

    # The cost of controlled swaps into working registers based on the k or
    # k - Q value.
    cs2 = 4 * n * (Nk - 1)

    # The QROM for the rotation angles the first time.
    cs2a = QR3(Nk * (M + n / 2), n * beta)[1] + QI(Nk * (M + n / 2))[1]

    # The QROM for the rotation angles the second time.
    cs2b = QR3(Nk * M, n * beta)[1] + QI(Nk * M)[1]

    # The cost of the rotations.
    cs3 = 16 * n * (beta - 2)

    # Cost for constructing contiguous register for outputting rotations.
    cs4 = (12 * Nk + 4 * np.ceil(np.log2(Nk * (M + n / 2))) +
           4 * np.ceil(np.log2(Nk * M)))

    # The cost of the controlled selection of the X vs Y.
    cs5 = 2 + 4 * (Nk - 1)

    # Cost of computing K-Q-G
    cs6 = 2 * nk

    # Cost of computing contiguous register.
    cs6 = cs6 + np.ceil(np.log2(Nkx)) * np.ceil(np.log2(Nky))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nky))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nky)) * np.ceil(np.log2(Nky))
    cs6 = cs6 + np.ceil(np.log2(Nk))
    cs6 = cs6 + np.ceil(np.log2(Nk)) * np.ceil(np.log2(Nkx))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nk))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nk)) * np.ceil(np.log2(Nky))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nky * Nk))
    cs6 = cs6 + np.ceil(np.log2(Nkx * Nky * Nk)) * np.ceil(np.log2(Nkz))
    cs6 = cs6 + np.ceil(np.log2(Nk**2))
    cs6 = cs6 + np.ceil(np.log2(Nkx**2)) * np.ceil(np.log2(M))
    cs6 = cs6 + np.ceil(np.log2(Nk**2 * M))

    # the times for is for computing and uncomputing twice
    cs6 = cs6 * 4

    # The QROM cost once we computed the contiguous register
    cs6 = cs6 + 2 * (QR3(Nk**2 * M, nk + chi)[1] + QI(Nk**2 * M)[1])

    # The remaining state preparation cost with coherent alias sampling.
    cs6 = cs6 + 4 * (nk + chi)

    # The costs of generating the symmetry in Q, with a +1 for the one-body
    # control.
    cs6 = cs6 + 4 * nk + 1

    # The total select cost.
    CS = cks + cka + cs1 + cs2 + cs2a + cs2b + cs3 + cs4 + cs5 + cs6

    # The reflection cost.
    costref = nc + 2 * nk + 3 * chi + 9

    cost = CPCP + CS + costref

    # Qubits for control for phase estimation
    ac1 = 2 * np.ceil(np.log2(iters + 1)) - 1

    # system qubits
    ac2 = n * Nk

    # various control qubits
    ac3 = nc + chi + nk + 10

    # phase gradient state
    ac4 = beta

    # T state
    ac5 = 1

    # kp = 2^QRa[d, m]
    kp = np.power(2, QR3(d, m)[0])

    # First round of QROM.
    ac12 = m * kp + np.ceil(np.log2(d / kp)) - 1

    # Temporary qubits from QROM.
    ac6 = m

    # First round of QROM.
    ac7 = m * (kp - 1) + np.ceil(np.log2(d / kp)) - 1

    # Qubit from inequality test for state preparation.
    ac8 = 1

    # Qubit from inequality test for mu, nu
    ac9 = 1

    # QROM for constructing superposition on k.
    ac10 = 2 * nk + 3 * 7

    # The contiguous register and k-Q
    ac11 = np.ceil(np.log2(Nk**2 * M) + nk)

    # output for k-state
    ac12 = nk + chi

    kn = np.power(2, QR3(Nk**2 * M, nk + chi)[0])

    ac13 = (kn - 1) * (nk + chi) + np.ceil(np.log2(Nk**2 * M)) - 1

    ac14 = chi + 1

    # The contiguous register.
    ac15 = np.ceil(np.log2(Nk * (M + n / 2)))

    kr = np.power(2, QR3(Nk * (M + n / 2), n * beta)[0])

    #
    ac16 = n * beta * kr + np.ceil(np.log2(Nk * (M + n / 2))) - 1

    # common ancilla costs
    aca = ac1 + ac2 + ac3 + ac4 + ac5 + ac6

    acc = np.max([ac13, ac14 + ac15 + ac16])

    acc = np.max([ac7, ac8 + ac9 + ac10 + ac11 + ac12 + acc])

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(aca + acc)

    # step-cost, Toffoli count, logical qubits
    return step_cost, total_cost, ancilla_cost



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from functools import reduce
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES
if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from ase.build import bulk

    from pyscf.pbc import gto, scf, mp
    from pyscf.pbc.tools import pyscf_ase

    from openfermion.resource_estimates.pbc.sf.compute_lambda_sf import (
        compute_lambda,)
    from openfermion.resource_estimates.pbc.sf.sf_integrals import (
        SingleFactorization,)
    from openfermion.resource_estimates.pbc.hamiltonian import (
        build_hamiltonian,
        cholesky_from_df_ints,
    )
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_lambda_calc():
    mf = make_diamond_113_szv()
    hcore, Luv = build_hamiltonian(mf)
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf)
    lambda_data = compute_lambda(hcore, helper)
    assert np.isclose(lambda_data.lambda_total, 2123.4342903006627)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_padding():
    ase_atom = bulk("H", "bcc", a=2.0, cubic=True)
    cell = gto.Cell()
    cell.atom = pyscf_ase.ase_atoms_to_pyscf(ase_atom)
    cell.a = ase_atom.cell[:].copy()
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.verbose = 0
    cell.build()

    kmesh = [1, 2, 2]
    kpts = cell.make_kpts(kmesh)
    nkpts = len(kpts)
    mf = scf.KRHF(cell, kpts).rs_density_fit()
    mf.with_df.mesh = cell.mesh
    mf.kernel()

    from pyscf.pbc.mp.kmp2 import _add_padding

    mymp = mp.KMP2(mf)
    Luv_padded = cholesky_from_df_ints(mymp)
    mo_coeff_padded = _add_padding(mymp, mymp.mo_coeff, mymp.mo_energy)[0]
    helper = SingleFactorization(cholesky_factor=Luv_padded, kmf=mf)
    assert mf.mo_coeff[0].shape[-1] != mo_coeff_padded[0].shape[-1]

    hcore_ao = mf.get_hcore()
    hcore_no_padding = np.asarray([
        reduce(np.dot, (mo.T.conj(), hcore_ao[k], mo))
        for k, mo in enumerate(mf.mo_coeff)
    ])
    hcore_padded = np.asarray([
        reduce(np.dot, (mo.T.conj(), hcore_ao[k], mo))
        for k, mo in enumerate(mo_coeff_padded)
    ])
    assert hcore_no_padding[0].shape != hcore_padded[0].shape
    assert np.isclose(np.sum(hcore_no_padding), np.sum(hcore_padded))
    Luv_no_padding = cholesky_from_df_ints(mymp, pad_mos_with_zeros=False)
    for k1 in range(nkpts):
        for k2 in range(nkpts):
            assert np.isclose(np.sum(Luv_padded[k1, k2]),
                              np.sum(Luv_no_padding[k1, k2]))

    helper_no_padding = SingleFactorization(cholesky_factor=Luv_no_padding,
                                            kmf=mf)
    lambda_data_pad = compute_lambda(hcore_no_padding, helper_no_padding)
    helper = SingleFactorization(cholesky_factor=Luv_padded, kmf=mf)
    lambda_data_no_pad = compute_lambda(hcore_padded, helper)
    assert np.isclose(lambda_data_pad.lambda_total,
                      lambda_data_no_pad.lambda_total)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    import jax
    import jax.numpy as jnp
    from pyscf.pbc import gto, mp, scf

    from openfermion.resource_estimates.pbc.hamiltonian import (
        build_momentum_transfer_mapping, cholesky_from_df_ints)
    from openfermion.resource_estimates.pbc.thc.factorizations.isdf import \
        solve_kmeans_kpisdf
    from openfermion.resource_estimates.pbc.thc.factorizations.thc_jax import (
        adagrad_opt_kpthc_batched, get_zeta_size, kpoint_thc_via_isdf,
        lbfgsb_opt_kpthc_l2reg, lbfgsb_opt_kpthc_l2reg_batched,
        make_contiguous_cholesky, pack_thc_factors,
        prepare_batched_data_indx_arrays, thc_objective_regularized,
        thc_objective_regularized_batched, unpack_thc_factors)
    from openfermion.resource_estimates.thc.utils.thc_factorization import \
        lbfgsb_opt_thc_l2reg
    from openfermion.resource_estimates.thc.utils.thc_factorization import \
        thc_objective_regularized as thc_obj_mol


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_kpoint_thc_reg_gamma():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.mesh = [11] * 3
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [1, 1, 1]
    kpts = cell.make_kpts(kmesh)
    num_kpts = len(kpts)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()
    num_mo = mf.mo_coeff[0].shape[-1]
    num_interp_points = 10 * mf.mo_coeff[0].shape[-1]
    kpt_thc = solve_kmeans_kpisdf(mf,
                                  num_interp_points,
                                  single_translation=False,
                                  verbose=False)
    chi, zeta, g_mapping = kpt_thc.chi, kpt_thc.zeta, kpt_thc.g_mapping
    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    buffer = np.zeros(2 * (chi.size + get_zeta_size(zeta)), dtype=np.float64)
    pack_thc_factors(chi, zeta, buffer)
    num_G_per_Q = [z.shape[0] for z in zeta]
    chi_unpacked, zeta_unpacked = unpack_thc_factors(buffer, num_interp_points,
                                                     num_mo, num_kpts,
                                                     num_G_per_Q)
    assert np.allclose(chi_unpacked, chi)
    for iq in range(num_kpts):
        assert np.allclose(zeta[iq], zeta_unpacked[iq])
    # force contiguous
    rsmf = scf.KRHF(mf.cell, mf.kpts).rs_density_fit()
    rsmf.verbose = 0
    rsmf.mo_occ = mf.mo_occ
    rsmf.mo_coeff = mf.mo_coeff
    rsmf.mo_energy = mf.mo_energy
    rsmf.with_df.mesh = mf.cell.mesh
    mymp = mp.KMP2(rsmf)
    Luv = cholesky_from_df_ints(mymp)
    Luv_cont = make_contiguous_cholesky(Luv)
    eri = np.einsum("npq,nrs->pqrs", Luv_cont[0, 0], Luv_cont[0, 0]).real
    buffer = np.zeros((chi.size + get_zeta_size(zeta)), dtype=np.float64)
    # transposed in openfermion
    buffer[:chi.size] = chi.T.real.ravel()
    buffer[chi.size:] = zeta[iq].real.ravel()
    np.random.seed(7)
    opt_param = lbfgsb_opt_thc_l2reg(
        eri,
        num_interp_points,
        maxiter=10,
        initial_guess=buffer,
        penalty_param=None,
    )
    chi_unpacked_mol = (opt_param[:chi.size].reshape(
        (num_interp_points, num_mo)).T)
    zeta_unpacked_mol = opt_param[chi.size:].reshape(zeta[0].shape)
    opt_param, _ = lbfgsb_opt_kpthc_l2reg(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        maxiter=10,
        penalty_param=None,
        disp_freq=-1,
    )
    chi_unpacked, zeta_unpacked = unpack_thc_factors(opt_param,
                                                     num_interp_points, num_mo,
                                                     num_kpts, num_G_per_Q)
    assert np.allclose(chi_unpacked[0], chi_unpacked_mol)
    assert np.allclose(zeta_unpacked[0], zeta_unpacked_mol)
    mol_obj = thc_obj_mol(buffer, num_mo, num_interp_points, eri, 1e-3)
    buffer = np.zeros(2 * (chi.size + get_zeta_size(zeta)), dtype=np.float64)
    pack_thc_factors(chi, zeta, buffer)
    gam_obj = thc_objective_regularized(
        buffer,
        num_mo,
        num_interp_points,
        momentum_map,
        g_mapping,
        Luv_cont,
        1e-3,
    )
    assert mol_obj - gam_obj < 1e-12


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_kpoint_thc_reg_batched():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.mesh = [11] * 3
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [1, 1, 3]
    kpts = cell.make_kpts(kmesh)
    num_kpts = len(kpts)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()
    momentum_map = build_momentum_transfer_mapping(cell, kpts)
    num_interp_points = 10 * cell.nao
    kpt_thc = solve_kmeans_kpisdf(mf,
                                  num_interp_points,
                                  single_translation=False,
                                  verbose=False)
    chi, zeta, g_mapping = kpt_thc.chi, kpt_thc.zeta, kpt_thc.g_mapping
    rsmf = scf.KRHF(mf.cell, mf.kpts).rs_density_fit()
    rsmf.mo_occ = mf.mo_occ
    rsmf.mo_coeff = mf.mo_coeff
    rsmf.mo_energy = mf.mo_energy
    rsmf.with_df.mesh = mf.cell.mesh
    mymp = mp.KMP2(rsmf)
    Luv = cholesky_from_df_ints(mymp)
    Luv_cont = make_contiguous_cholesky(Luv)
    buffer = np.zeros(2 * (chi.size + get_zeta_size(zeta)), dtype=np.float64)
    # Pack THC factors into flat array
    pack_thc_factors(chi, zeta, buffer)
    num_G_per_Q = [z.shape[0] for z in zeta]
    num_mo = mf.mo_coeff[0].shape[-1]
    chi_unpacked, zeta_unpacked = unpack_thc_factors(buffer, num_interp_points,
                                                     num_mo, num_kpts,
                                                     num_G_per_Q)
    # Test packing/unpacking operation
    assert np.allclose(chi_unpacked, chi)
    for iq in range(num_kpts):
        assert np.allclose(zeta[iq], zeta_unpacked[iq])
    # Test objective is the same batched/non-batched
    penalty = 1e-3
    obj_ref = thc_objective_regularized(
        buffer,
        num_mo,
        num_interp_points,
        momentum_map,
        g_mapping,
        Luv_cont,
        penalty,
    )
    # # Test gradient is the same
    indx_arrays = prepare_batched_data_indx_arrays(momentum_map, g_mapping)
    batch_size = num_kpts**2
    obj_batched = thc_objective_regularized_batched(
        buffer,
        num_mo,
        num_interp_points,
        momentum_map,
        g_mapping,
        Luv_cont,
        indx_arrays,
        batch_size,
        penalty_param=penalty,
    )
    assert abs(obj_ref - obj_batched) < 1e-12
    grad_ref_fun = jax.grad(thc_objective_regularized)
    grad_ref = grad_ref_fun(
        buffer,
        num_mo,
        num_interp_points,
        momentum_map,
        g_mapping,
        Luv_cont,
        penalty,
    )
    # Test gradient is the same
    grad_batched_fun = jax.grad(thc_objective_regularized_batched)
    grad_batched = grad_batched_fun(
        buffer,
        num_mo,
        num_interp_points,
        momentum_map,
        g_mapping,
        Luv_cont,
        indx_arrays,
        batch_size,
        penalty,
    )
    assert np.allclose(grad_batched, grad_ref)
    opt_param, _ = lbfgsb_opt_kpthc_l2reg(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        maxiter=2,
        penalty_param=1e-3,
        disp_freq=-1,
    )
    opt_param_batched, _ = lbfgsb_opt_kpthc_l2reg_batched(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        maxiter=2,
        penalty_param=1e-3,
        disp_freq=-1,
    )
    assert np.allclose(opt_param, opt_param_batched)
    batch_size = 7
    opt_param_batched_diff_batch, _ = lbfgsb_opt_kpthc_l2reg_batched(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        batch_size=batch_size,
        maxiter=2,
        penalty_param=1e-3,
        disp_freq=-1,
    )
    assert np.allclose(opt_param_batched, opt_param_batched_diff_batch)
    ada_param, _ = adagrad_opt_kpthc_batched(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        maxiter=2,
        batch_size=1,
    )
    assert np.allclose(opt_param, opt_param_batched)
    batch_size = 7
    ada_param_diff_batch, _ = adagrad_opt_kpthc_batched(
        chi,
        zeta,
        momentum_map,
        g_mapping,
        jnp.array(Luv_cont),
        batch_size=batch_size,
        maxiter=2,
    )
    assert np.allclose(ada_param, ada_param_diff_batch)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_kpoint_thc_helper():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.mesh = [11] * 3
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [1, 1, 2]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts)
    mf.kernel()
    rsmf = scf.KRHF(mf.cell, mf.kpts).rs_density_fit()
    rsmf.verbose = 0
    rsmf.mo_occ = mf.mo_occ
    rsmf.mo_coeff = mf.mo_coeff
    rsmf.mo_energy = mf.mo_energy
    rsmf.with_df.mesh = mf.cell.mesh
    mymp = mp.KMP2(rsmf)
    Luv = cholesky_from_df_ints(mymp)
    cthc = 5
    num_mo = mf.mo_coeff[0].shape[-1]
    # Just testing function runs
    kpt_thc, _ = kpoint_thc_via_isdf(
        mf,
        Luv,
        cthc * num_mo,
        perform_adagrad_opt=False,
        perform_bfgs_opt=False,
    )
    kpt_thc_bfgs, _ = kpoint_thc_via_isdf(
        mf,
        Luv,
        cthc * num_mo,
        perform_adagrad_opt=False,
        perform_bfgs_opt=True,
        bfgs_maxiter=10,
        initial_guess=kpt_thc,
    )
    kpoint_thc_via_isdf(
        mf,
        Luv,
        cthc * num_mo,
        perform_adagrad_opt=True,
        perform_bfgs_opt=True,
        bfgs_maxiter=10,
        adagrad_maxiter=10,
        initial_guess=kpt_thc_bfgs,
    )



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Methods for overwriting coupled cluster single and doubles (CCSD) eris."""

from pyscf.lib import logger
from pyscf.pbc.lib.kpts_helper import loop_kkk
from pyscf.pbc.cc.kccsd_uhf import _make_eris_incore
from pyscf.pbc.cc.kccsd_rhf import _ERIS
from pyscf.pbc import cc, scf


def build_cc_inst(pyscf_mf):
    """Build PBC CC instance.

    If ROHF build KUCCSD object

    Args:
        pyscf_mf: pyscf mean field object (RHF or ROHF).

    Returns:
        cc_inst: Coupled cluster instance (RHF->KRCCSD, ROHF->KUCCSD).
    """
    if pyscf_mf.cell.spin == 0:
        cc_inst = cc.KRCCSD(pyscf_mf)
    else:
        u_from_ro = scf.addons.convert_to_uhf(pyscf_mf)
        cc_inst = cc.KUCCSD(u_from_ro)
    return cc_inst


def build_approximate_eris(krcc_inst, eri_helper, eris=None):
    """Update coupled cluster eris object with approximate integrals.

    Arguments:
        cc: pyscf PBC KRCCSD object.
        eri_helper: Approximate ERIs helper function which defines MO integrals.
        eris: pyscf _ERIS object. Optional, if present overwrite this eris
            object rather than build from scratch.

    Returns:
        eris: pyscf _ERIS object updated to hold approximate eris
            defined by eri_helper.
    """
    log = logger.Logger(krcc_inst.stdout, krcc_inst.verbose)
    kconserv = krcc_inst.khelper.kconserv
    khelper = krcc_inst.khelper
    nocc = krcc_inst.nocc
    nkpts = krcc_inst.nkpts
    dtype = krcc_inst.mo_coeff[0].dtype
    if eris is not None:
        log.info("Modifying coupled cluster _ERIS object inplace using "
                 f"{eri_helper.__class__}.")
        out_eris = eris
    else:
        log.info(f"Rebuilding coupled cluster _ERIS object using "
                 " {eri_helper.__class__}.")
        out_eris = _ERIS(krcc_inst)
    for ikp, ikq, ikr in khelper.symm_map.keys():
        iks = kconserv[ikp, ikq, ikr]
        kpts = [ikp, ikq, ikr, iks]
        eri_kpt = eri_helper.get_eri(kpts) / nkpts
        if dtype == float:
            eri_kpt = eri_kpt.real
        eri_kpt = eri_kpt
        for kp, kq, kr in khelper.symm_map[(ikp, ikq, ikr)]:
            eri_kpt_symm = khelper.transform_symm(eri_kpt, kp, kq,
                                                  kr).transpose(0, 2, 1, 3)
            out_eris.oooo[kp, kr, kq] = eri_kpt_symm[:nocc, :nocc, :nocc, :nocc]
            out_eris.ooov[kp, kr, kq] = eri_kpt_symm[:nocc, :nocc, :nocc, nocc:]
            out_eris.oovv[kp, kr, kq] = eri_kpt_symm[:nocc, :nocc, nocc:, nocc:]
            out_eris.ovov[kp, kr, kq] = eri_kpt_symm[:nocc, nocc:, :nocc, nocc:]
            out_eris.voov[kp, kr, kq] = eri_kpt_symm[nocc:, :nocc, :nocc, nocc:]
            out_eris.vovv[kp, kr, kq] = eri_kpt_symm[nocc:, :nocc, nocc:, nocc:]
            out_eris.vvvv[kp, kr, kq] = eri_kpt_symm[nocc:, nocc:, nocc:, nocc:]
    return out_eris


def build_approximate_eris_rohf(kucc_inst, eri_helper, eris=None):
    """Update unrestricted coupled cluster eris object with approximate ERIs.

    KROCCSD is run through KUCCSD object, but we expect (and build) RO
    integrals only.

    Arguments:
        kucc_inst: pyscf PBC KUCCSD object. Only ROHF integrals are supported.
        eri_helper: Approximate ERIs helper function which defines MO integrals.
        eris: pyscf _ERIS object. Optional, if present overwrite this eris
            object rather than build from scratch.

    Returns:
        eris: pyscf _ERIS object updated to hold approximate eris defined by
            eri_helper.
    """
    log = logger.Logger(kucc_inst.stdout, kucc_inst.verbose)
    kconserv = kucc_inst.khelper.kconserv
    nocca, noccb = kucc_inst.nocc
    nkpts = kucc_inst.nkpts
    if eris is not None:
        log.info("Modifying coupled cluster _ERIS object inplace using "
                 f"{eri_helper.__class__}.")
        out_eris = eris
    else:
        log.info("Rebuilding coupled cluster _ERIS object using "
                 f"{eri_helper.__class__}.")
        out_eris = _make_eris_incore(kucc_inst)
    for kp, kq, kr in loop_kkk(nkpts):
        ks = kconserv[kp, kq, kr]
        kpts = [kp, kq, kr, ks]
        tmp = eri_helper.get_eri(kpts) / nkpts
        out_eris.oooo[kp, kq, kr] = tmp[:nocca, :nocca, :nocca, :nocca]
        out_eris.ooov[kp, kq, kr] = tmp[:nocca, :nocca, :nocca, nocca:]
        out_eris.oovv[kp, kq, kr] = tmp[:nocca, :nocca, nocca:, nocca:]
        out_eris.ovov[kp, kq, kr] = tmp[:nocca, nocca:, :nocca, nocca:]
        out_eris.voov[kq, kp, ks] = (
            tmp[:nocca, nocca:, nocca:, :nocca].conj().transpose(1, 0, 3, 2))
        out_eris.vovv[kq, kp, ks] = (
            tmp[:nocca, nocca:, nocca:, nocca:].conj().transpose(1, 0, 3, 2))

    for kp, kq, kr in loop_kkk(nkpts):
        ks = kconserv[kp, kq, kr]
        kpts = [kp, kq, kr, ks]
        tmp = eri_helper.get_eri(kpts) / nkpts
        out_eris.OOOO[kp, kq, kr] = tmp[:noccb, :noccb, :noccb, :noccb]
        out_eris.OOOV[kp, kq, kr] = tmp[:noccb, :noccb, :noccb, noccb:]
        out_eris.OOVV[kp, kq, kr] = tmp[:noccb, :noccb, noccb:, noccb:]
        out_eris.OVOV[kp, kq, kr] = tmp[:noccb, noccb:, :noccb, noccb:]
        out_eris.VOOV[kq, kp, ks] = (
            tmp[:noccb, noccb:, noccb:, :noccb].conj().transpose(1, 0, 3, 2))
        out_eris.VOVV[kq, kp, ks] = (
            tmp[:noccb, noccb:, noccb:, noccb:].conj().transpose(1, 0, 3, 2))

    for kp, kq, kr in loop_kkk(nkpts):
        ks = kconserv[kp, kq, kr]
        kpts = [kp, kq, kr, ks]
        tmp = eri_helper.get_eri(kpts) / nkpts
        out_eris.ooOO[kp, kq, kr] = tmp[:nocca, :nocca, :noccb, :noccb]
        out_eris.ooOV[kp, kq, kr] = tmp[:nocca, :nocca, :noccb, noccb:]
        out_eris.ooVV[kp, kq, kr] = tmp[:nocca, :nocca, noccb:, noccb:]
        out_eris.ovOV[kp, kq, kr] = tmp[:nocca, nocca:, :noccb, noccb:]
        out_eris.voOV[kq, kp, ks] = (
            tmp[:nocca, nocca:, noccb:, :noccb].conj().transpose(1, 0, 3, 2))
        out_eris.voVV[kq, kp, ks] = (
            tmp[:nocca, nocca:, noccb:, noccb:].conj().transpose(1, 0, 3, 2))

    for kp, kq, kr in loop_kkk(nkpts):
        ks = kconserv[kp, kq, kr]
        kpts = [kp, kq, kr, ks]
        tmp = eri_helper.get_eri(kpts) / nkpts
        # out_eris.OOoo[kp,kq,kr] = tmp[:noccb,:noccb,:nocca,:nocca]
        out_eris.OOov[kp, kq, kr] = tmp[:noccb, :noccb, :nocca, nocca:]
        out_eris.OOvv[kp, kq, kr] = tmp[:noccb, :noccb, nocca:, nocca:]
        out_eris.OVov[kp, kq, kr] = tmp[:noccb, noccb:, :nocca, nocca:]
        out_eris.VOov[kq, kp, ks] = (
            tmp[:noccb, noccb:, nocca:, :nocca].conj().transpose(1, 0, 3, 2))
        out_eris.VOvv[kq, kp, ks] = (
            tmp[:noccb, noccb:, nocca:, nocca:].conj().transpose(1, 0, 3, 2))
    # Force CCSD to use eri tensors.
    out_eris.Lpv = None
    out_eris.LPV = None

    return out_eris


def compute_emp2_approx(mf, intgl_helper) -> float:
    """Compute MP2 energy given an integral helper

    Args:
        mf: pyscf MF object (RHF or ROHF).
        ingl_helper: Integral helper (sparse, SF, DF, or THC)

    Returns:
        emp: MP2 total energy.
    """
    cc_inst = build_cc_inst(mf)
    approx_eris = build_approximate_eris(cc_inst, intgl_helper)
    emp2_approx, _, _ = cc_inst.init_amps(approx_eris)
    emp2_approx += mf.e_tot
    return emp2_approx



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from functools import reduce

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import mp
    from openfermion.resource_estimates.pbc.sparse.\
        compute_lambda_sparse import compute_lambda
    from openfermion.resource_estimates.pbc.sparse.sparse_integrals import (
        SparseFactorization,)
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_lambda_sparse():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    Luv = cholesky_from_df_ints(mymp)
    helper = SparseFactorization(cholesky_factor=Luv, kmf=mf)

    hcore_ao = mf.get_hcore()
    hcore_mo = np.asarray([
        reduce(np.dot, (mo.T.conj(), hcore_ao[k], mo))
        for k, mo in enumerate(mf.mo_coeff)
    ])
    compute_lambda(hcore_mo, helper)



#coverage:ignore
"""Test cases for util.py
"""
from openfermion.resource_estimates.utils import QR, QI, QR2, QI2, power_two


def test_QR():
    """ Tests function QR which gives the minimum cost for a QROM over L values
        of size M.
    """
    # Tests checked against Mathematica noteboook `costingTHC.nb`
    # Arguments are otherwise random
    assert QR(12341234, 5670) == (6, 550042)
    assert QR(12201990, 520199) == (2, 4611095)


def test_QI():
    """ Tests function QI which gives the minimum cost for inverse QROM over
        L values.
    """
    # Tests checked against Mathematica noteboook `costingTHC.nb`
    # Arguments are otherwise random
    assert QI(987654) == (10, 1989)
    assert QI(8052021) == (11, 5980)


def test_QR2():
    """ Tests function QR2 which gives the minimum cost for a QROM with two
        registers.
    """
    # Tests checked against Mathematica noteboook `costingsf.nb`
    # Arguments are otherwise random
    assert QR2(12, 34, 81) == (2, 2, 345)
    assert QR2(712, 340111, 72345) == (4, 16, 8341481)


def test_QI2():
    """ Tests function QI which gives the minimum cost for inverse QROM with
        two registers.
    """
    # Tests checked against Mathematica noteboook `costingsf.nb`
    # Arguments are otherwise random
    assert QI2(1234, 5678) == (32, 64, 5519)
    assert QI2(7120, 1340111) == (4, 32768, 204052)


def test_power_two():
    """ Test for power_two(m) which returns power of 2 that is a factor of m """
    try:
        power_two(-1234)
    except AssertionError:
        pass
    assert power_two(0) == 0
    assert power_two(2) == 1
    assert power_two(3) == 0
    assert power_two(104) == 3  # 2**3 * 13
    assert power_two(128) == 7  # 2**7
    assert power_two(393120) == 5  # 2**5 * 3**3 * 5 * 7 * 13



#coverage:ignore
"""Test cases for costing_df.py
"""
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates import df


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_reiher_df():
    """ Reproduce Reiher et al orbital DF FT costs from paper """
    DE = 0.001
    CHI = 10

    # Reiher et al orbitals
    N = 108
    LAM = 294.8
    L = 360
    LXI = 13031
    BETA = 16

    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps.
    # Then we input that into the function again.
    output = df.compute_cost(N, LAM, DE, L, LXI, CHI, BETA, stps=20000)
    stps1 = output[0]
    output = df.compute_cost(N, LAM, DE, L, LXI, CHI, BETA, stps1)
    assert output == (21753, 10073183463, 3725)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_li_df():
    """ Reproduce Li et al orbital DF FT costs from paper """
    DE = 0.001
    CHI = 10

    # Li et al orbitals
    N = 152
    LAM = 1171.2
    L = 394
    LXI = 20115
    BETA = 20

    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps.
    # Then we input that into the function again.
    output = df.compute_cost(N, LAM, DE, L, LXI, CHI, BETA, stps=20000)
    stps2 = output[0]
    output = df.compute_cost(N, LAM, DE, L, LXI, CHI, BETA, stps2)
    assert output == (35008, 64404812736, 6404)



#coverage:ignore
"""Test cases for pyscf_utils.py
"""
import unittest
from os import path

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf import cc, gto, scf

    from openfermion.resource_estimates import df, sf
    from openfermion.resource_estimates.molecule import (
        ccsd_t, factorized_ccsd_t, load_casfile_to_pyscf, open_shell_t1_d1,
        pyscf_to_cas, stability)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
class OpenFermionPyscfUtilsTest(unittest.TestCase):

    def test_full_ccsd_t(self):
        """ Test resource_estimates full CCSD(T) from h1/eri/ecore tensors
            matches regular PySCF CCSD(T)
        """

        for scf_type in ['rhf', 'rohf']:
            mol = gto.Mole()
            mol.atom = 'H 0 0 0; F 0 0 1.1'
            mol.charge = 0
            if scf_type == 'rhf':
                mol.spin = 0
            elif scf_type == 'rohf':
                mol.spin = 2
            mol.basis = 'ccpvtz'
            mol.symmetry = False
            mol.build()

            if scf_type == 'rhf':
                mf = scf.RHF(mol)
            elif scf_type == 'rohf':
                mf = scf.ROHF(mol)

            mf.init_guess = 'mindo'
            mf.conv_tol = 1e-10
            mf.kernel()
            mf = stability(mf)

            # Do PySCF CCSD(T)
            mycc = cc.CCSD(mf)
            mycc.max_cycle = 500
            mycc.conv_tol = 1E-9
            mycc.conv_tol_normt = 1E-5
            mycc.diis_space = 24
            mycc.diis_start_cycle = 4
            mycc.kernel()
            et = mycc.ccsd_t()

            pyscf_escf = mf.e_tot
            pyscf_ecor = mycc.e_corr + et
            pyscf_etot = pyscf_escf + pyscf_ecor
            pyscf_results = np.array([pyscf_escf, pyscf_ecor, pyscf_etot])

            n_elec = mol.nelectron
            n_orb = mf.mo_coeff[0].shape[-1]

            resource_estimates_results = ccsd_t(
                *pyscf_to_cas(mf, n_orb, n_elec))
            resource_estimates_results = np.asarray(resource_estimates_results)

            # ignore relative tolerance, we just want absolute tolerance
            assert np.allclose(pyscf_results,
                               resource_estimates_results,
                               rtol=1E-14)

    def test_reduced_ccsd_t(self):
        """ Test resource_estimates reduced (2e space) CCSD(T) from tensors
            matches PySCF CAS(2e,No)
        """

        for scf_type in ['rhf', 'rohf']:
            mol = gto.Mole()
            mol.atom = 'H 0 0 0; F 0 0 1.1'
            mol.charge = 0
            if scf_type == 'rhf':
                mol.spin = 0
            elif scf_type == 'rohf':
                mol.spin = 2
            mol.basis = 'ccpvtz'
            mol.symmetry = False
            mol.build()

            if scf_type == 'rhf':
                mf = scf.RHF(mol)
            elif scf_type == 'rohf':
                mf = scf.ROHF(mol)

            mf.init_guess = 'mindo'
            mf.conv_tol = 1e-10
            mf.kernel()
            mf = stability(mf)

            # PySCF CAS(No,2e) for 2 electrons CCSD (and so CCSD(T)) is exact
            n_elec = 2  # electrons
            n_orb = mf.mo_coeff[0].shape[-1] - mf.mol.nelectron - n_elec
            mycas = mf.CASCI(n_orb, n_elec).run()

            pyscf_etot = mycas.e_tot

            # Don't do triples (it's zero anyway for 2e) b/c div by zero w/ ROHF
            _, _, resource_estimates_etot = ccsd_t(
                *pyscf_to_cas(mf, n_orb, n_elec), no_triples=True)

            # ignore relative tolerance, we just want absolute tolerance
            assert np.isclose(pyscf_etot, resource_estimates_etot, rtol=1E-14)

    def test_reiher_sf_ccsd_t(self):
        """ Reproduce Reiher et al FeMoco SF CCSD(T) errors from paper """

        NAME = path.join(path.dirname(__file__), '../integrals/eri_reiher.h5')
        _, mf = load_casfile_to_pyscf(NAME, num_alpha=27, num_beta=27)
        _, ecorr, _ = factorized_ccsd_t(
            mf, eri_rr=None)  # use full (local) ERIs for 2-body
        exact_energy = ecorr
        rank = 100
        eri_rr, _ = sf.factorize(mf._eri, rank)
        _, ecorr, _ = factorized_ccsd_t(mf, eri_rr)
        appx_energy = ecorr

        error = (appx_energy - exact_energy) * 1E3  # mEh

        assert np.isclose(np.round(error, decimals=2), 1.55)

    def test_reiher_df_ccsd_t(self):
        """ Reproduce Reiher et al FeMoco DF CCSD(T) errors from paper """

        NAME = path.join(path.dirname(__file__), '../integrals/eri_reiher.h5')
        _, mf = load_casfile_to_pyscf(NAME, num_alpha=27, num_beta=27)
        _, ecorr, _ = factorized_ccsd_t(
            mf, eri_rr=None)  # use full (local) ERIs for 2-body
        exact_energy = ecorr
        appx_energy = []
        THRESH = 0.00125
        eri_rr, _, _, _ = df.factorize(mf._eri, thresh=THRESH)
        _, ecorr, _ = factorized_ccsd_t(mf, eri_rr)
        appx_energy = ecorr

        error = (appx_energy - exact_energy) * 1E3  # mEh

        assert np.isclose(np.round(error, decimals=2), 0.44)

    def test_t1_d1_openshell(self):
        """Test open shell t1-diagnostic by reducing back to closed shell"""
        mol = gto.M()
        mol.atom = 'N 0 0 0; N 0 0 1.4'
        mol.basis = 'cc-pvtz'
        mol.spin = 0
        mol.build()

        mf = scf.RHF(mol)
        mf.kernel()

        mycc = cc.CCSD(mf)
        mycc.kernel()

        true_t1d, true_d1d = mycc.get_t1_diagnostic(), mycc.get_d1_diagnostic()

        uhf_mf = scf.convert_to_uhf(mf)
        mycc_uhf = cc.CCSD(uhf_mf)
        mycc_uhf.kernel()
        t1a, t1b = mycc_uhf.t1
        test_t1d, test_d1d = open_shell_t1_d1(
            t1a, t1b, uhf_mf.mo_occ[0] + uhf_mf.mo_occ[1], uhf_mf.nelec[0],
            uhf_mf.nelec[1])

        assert np.isclose(test_t1d, true_t1d)
        assert np.isclose(test_d1d, true_d1d)
        assert np.sqrt(2) * test_t1d <= test_d1d

    def test_t1_d1_oxygen(self):
        """Test open shell t1-diagnostic on O2 molecule

        Compare with output from Psi4

        * Input:

        molecule oxygen {
          0 3
          O 0.0 0.0 0.0
          O 0.0 0.0 1.1
          no_reorient
          symmetry c1
        }

        set {
          reference rohf
          basis cc-pvtz
        }

        energy('CCSD')

        * Output
        @ROHF Final Energy:  -149.65170765644311

                       Solving CC Amplitude Equations
                        ------------------------------
        Iter     Energy           RMS        T1Diag      D1Diag    New D1Diag
        ----     ----------    ---------   ----------  ----------  ----------
        ...
        10        -0.464506    1.907e-07    0.004390    0.009077    0.009077
        11        -0.464506    5.104e-08    0.004390    0.009077    0.009077

        Iterations converged.
        """

        mol = gto.M()
        mol.atom = 'O 0 0 0; O 0 0 1.1'
        mol.basis = 'cc-pvtz'
        mol.spin = 2
        mol.build()

        mf = scf.ROHF(mol)
        mf.kernel()

        uhf_mf = scf.convert_to_uhf(mf)
        mycc_uhf = cc.CCSD(mf)
        mycc_uhf.kernel()

        t1a, t1b = mycc_uhf.t1
        test_t1d, test_d1d = open_shell_t1_d1(
            t1a, t1b, uhf_mf.mo_occ[0] + uhf_mf.mo_occ[1], uhf_mf.nelec[0],
            uhf_mf.nelec[1])

        assert np.isclose(mf.e_tot, -149.651708, atol=1e-6)
        assert np.isclose(mycc_uhf.e_corr, -0.464507, atol=1e-6)
        assert np.isclose(test_t1d, 0.004390, atol=1e-4)
        assert np.isclose(test_d1d, 0.009077, atol=1e-4)

    def test_t1_d1_bound(self):
        """sqrt(2) * t1 <= d1"""
        mol = gto.M()
        mol.atom = 'O 0 0 0; O 0 0 1.4'
        mol.basis = 'cc-pvtz'
        mol.spin = 2
        mol.build()
        mf = scf.ROHF(mol)
        mf.kernel()
        mycc = cc.CCSD(mf)
        mycc.kernel()
        uhf_mf = scf.convert_to_uhf(mf)
        mycc_uhf = cc.CCSD(uhf_mf)
        mycc_uhf.kernel()
        t1a, t1b = mycc_uhf.t1
        test_t1d, test_d1d = open_shell_t1_d1(
            t1a, t1b, uhf_mf.mo_occ[0] + uhf_mf.mo_occ[1], uhf_mf.nelec[0],
            uhf_mf.nelec[1])
        assert np.sqrt(2) * test_t1d <= test_d1d



#coverage:ignore
"""Test cases for compute_lambda_sf.py
"""
from os import path

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES, sf

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.molecule import load_casfile_to_pyscf


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_reiher_sf_lambda():
    """ Reproduce Reiher et al orbital SF lambda from paper """

    RANK = 200

    NAME = path.join(path.dirname(__file__), '../integrals/eri_reiher.h5')
    _, reiher_mf = load_casfile_to_pyscf(NAME, num_alpha=27, num_beta=27)
    eri_rr, sf_factors = sf.factorize(reiher_mf._eri, RANK)
    lambda_tot = sf.compute_lambda(reiher_mf, sf_factors)
    assert eri_rr.shape[0] * 2 == 108
    assert np.isclose(lambda_tot, 4258.0)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module constructs Hamiltonians of the Richardson Gaudin type.
"""
from itertools import chain, product
import numpy
from openfermion.ops.representations import (PolynomialTensor,
                                             get_tensors_from_integrals)
from openfermion.ops.representations import DOCIHamiltonian
from openfermion.ops import QubitOperator


class RichardsonGaudin(DOCIHamiltonian):
    r"""Richardson Gaudin model.

    Class for storing and constructing Richardson Gaudin hamiltonians
    combining an equi-distant potential ladder like potential per
    qubit with a uniform coupling between any pair of
    qubits with coupling strength g, which can be either attractive
    (g<0) or repulsive (g>0).

    The operators represented by this class has the form:

    .. math::

        H = \sum_{p=0} (p + 1) N_p + g/2 \sum_{p < q} P_p^\dagger P_q,

    where

    .. math::

        \begin{align}
        N_p &= (1 - \sigma^Z_p)/2, \\
        P_p &= a_{p,\beta} a_{p,\alpha} = S^{-} = \sigma^X + i \sigma^Y, \\
        g &= constant coupling term
        \end{align}

    Note;
        The diagonal of the Hamiltonian is composed of the values in
        range((n_qubits+1)*n_qubits//2+1).
    """

    def __init__(self, g, n_qubits):
        r"""Richardson Gaudin model on a given number of qubits.

        Args:
            g (float): Coupling strength
            n_qubits (int): Number of qubits
        """
        hc = numpy.zeros((n_qubits,))
        hr1 = numpy.zeros((n_qubits, n_qubits))
        hr2 = numpy.zeros((n_qubits, n_qubits))
        for p in range(n_qubits):
            hc[p] = 2 * (p + 1)
            for q in range(n_qubits):
                if p != q:
                    hr1[p, q] = g
        super().__init__(0, hc, hr1, hr2)

    @DOCIHamiltonian.constant.setter
    def constant(self, value):
        raise TypeError('Raw edits of the constant of a RichardsonGaudin model'
                        'is not allowed. Either adjust the g parameter '
                        'or cast to another PolynomialTensor class.')

    @DOCIHamiltonian.n_body_tensors.setter
    def n_body_tensors(self, value):
        raise TypeError(
            'Raw edits of the n_body_tensors of a RichardsonGaudin model'
            'is not allowed. Either adjust the g parameter '
            'or cast to another PolynomialTensor class.')

    def get_antisymmetrized_tensors(self):
        r"""Antisymmetrized Tensors
        Directly returns antisymmetrized tensors, which, when used
        to construct an FermionOperator via an InteractionOperator
        produce a FermionOperator that acts like this RichardsonGaudin
        Hamiltonian on the paired (seniority zero) subspace.
        Compared to the FermionOperator that can be obtained via the
        n_body_tensors property from the DOCIHamiltonian class
        the FermionOperator from the tensors returned by this function
        do not contain same spin coupling terms. These terms
        act trivially on the paired subspace and this the two Hamiltonian
        agree on any senioirty zero state.
        Returns:
            tuple: Tuple of one and two body tensors.
        """
        g = self.hr1[0, 1]
        spatial_orbs = self.hc.shape[0]
        h1 = numpy.diag(numpy.arange(spatial_orbs) + 1)
        h1 = numpy.kron(h1, numpy.eye(2))
        h2 = numpy.zeros((2 * spatial_orbs,) * 4)
        for p, q in product(range(spatial_orbs), repeat=2):
            if p != q:
                h2[2 * p, 2 * p + 1, 2 * q + 1, 2 * q] = g / 2
                h2[2 * p + 1, 2 * p, 2 * q, 2 * q + 1] = g / 2

        h2 = h2 - numpy.einsum('ijlk', h2)

        return h1, h2



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from pyscf.pbc import scf
from pyscf.pbc.tools.k2gamma import kpts_to_kmesh

from openfermion.resource_estimates.pbc.thc.factorizations.thc_jax import (
    kpoint_thc_via_isdf,)
from openfermion.resource_estimates.pbc.resources.data_types import PBCResources
from openfermion.resource_estimates.pbc.thc.thc_integrals import (
    KPTHCDoubleTranslation,)
from openfermion.resource_estimates.pbc.hamiltonian.cc_extensions import (
    build_approximate_eris,
    build_cc_inst,
    build_approximate_eris_rohf,
)
from openfermion.resource_estimates.pbc.hamiltonian import (
    build_hamiltonian,)
from openfermion.resource_estimates.pbc.thc.compute_lambda_thc import (
    compute_lambda,)
from openfermion.resource_estimates.pbc.thc.compute_thc_resources import (
    compute_cost,)


@dataclass
class THCResources(PBCResources):
    beta: int = 20


def generate_costing_table(
        pyscf_mf: scf.HF,
        thc_rank_params: Union[list, npt.NDArray],
        name="pbc",
        chi: int = 10,
        beta: int = 20,
        dE_for_qpe: float = 0.0016,
        reoptimize: bool = True,
        bfgs_maxiter: int = 3000,
        adagrad_maxiter: int = 3000,
        fft_df_mesh: Union[None, list] = None,
        energy_method: str = "MP2",
) -> pd.DataFrame:
    """Generate resource estimate costing table for THC Hamiltonian.

    Arguments:
        pyscf_mf: k-point pyscf mean-field object
        thc_rank_params: Array of (integer) auxiliary index cutoff values
        name: Optional descriptive name for simulation.
        chi: the number of bits for the representation of the coefficients
        beta: the number of bits for rotations.
        dE_for_qpe: Phase estimation epsilon.
        reoptimize: Whether or not to perform regularized reoptimization of THC
            factors.
        bfgs_maxiter: Max number of BFGS steps.
        adagrad_maxiter: Max number of AdaGrad steps.
        fft_df_mesh: FFTDF mesh for ISDF.
        energy_method: Method to determine energy with (CCSD or MP2.)
    Returns
        resources: Table of resource estimates.
    """
    kmesh = kpts_to_kmesh(pyscf_mf.cell, pyscf_mf.kpts)
    cc_inst = build_cc_inst(pyscf_mf)
    exact_eris = cc_inst.ao2mo()
    if energy_method.lower() == "mp2":
        energy_function = lambda x: cc_inst.init_amps(x)
        reference_energy, _, _ = energy_function(exact_eris)
    elif energy_method.lower() == "ccsd":
        energy_function = lambda x: cc_inst.kernel(eris=x)
        reference_energy, _, _ = energy_function(exact_eris)
    else:
        raise ValueError(f"Unknown value for energy_method: {energy_method}")

    hcore, chol = build_hamiltonian(pyscf_mf)
    num_spin_orbs = 2 * hcore[0].shape[-1]
    num_kpts = np.prod(kmesh)
    thc_resource_obj = THCResources(
        system_name=name,
        num_spin_orbitals=num_spin_orbs,
        num_kpts=num_kpts,
        dE=dE_for_qpe,
        chi=chi,
        beta=beta,
        energy_method=energy_method,
        exact_energy=np.real(reference_energy),
    )

    # For the ISDF guess we need an FFTDF MF object (really just need the grids
    # so a bit of a hack) Since we're fitting to RSGDF it isn't very important
    # what the value of the FFT mesh is, and there is some tradeoff (grid
    # density vs comp time) which is not critical (see isdf notebook).
    # Subsequent optimization attempts to fit to the RSGDF integrals which
    # should hopefully be somewhat close to the FFTDF ones for ISDF to be a good
    # starting point.
    mf_fftdf = scf.KRHF(pyscf_mf.cell, pyscf_mf.kpts)
    mf_fftdf.max_memory = 180000
    mf_fftdf.kpts = pyscf_mf.kpts
    mf_fftdf.e_tot = pyscf_mf.e_tot
    mf_fftdf.mo_coeff = pyscf_mf.mo_coeff
    mf_fftdf.mo_energy = pyscf_mf.mo_energy
    mf_fftdf.mo_occ = pyscf_mf.mo_occ
    if fft_df_mesh is not None:
        mf_fftdf.with_df.mesh = fft_df_mesh
    approx_eris = exact_eris
    for thc_rank in thc_rank_params:
        num_thc = thc_rank * num_spin_orbs // 2
        kpt_thc, _ = kpoint_thc_via_isdf(
            mf_fftdf,
            chol,
            num_thc,
            perform_adagrad_opt=reoptimize,
            perform_bfgs_opt=reoptimize,
            bfgs_maxiter=bfgs_maxiter,
            adagrad_maxiter=adagrad_maxiter,
        )
        thc_helper = KPTHCDoubleTranslation(kpt_thc.chi,
                                            kpt_thc.zeta,
                                            pyscf_mf,
                                            chol=chol)
        thc_lambda = compute_lambda(hcore, thc_helper)
        kmesh = kpts_to_kmesh(pyscf_mf.cell, pyscf_mf.kpts)
        if pyscf_mf.cell.spin == 0:
            approx_eris = build_approximate_eris(cc_inst,
                                                 thc_helper,
                                                 eris=approx_eris)
        else:
            approx_eris = build_approximate_eris_rohf(cc_inst,
                                                      thc_helper,
                                                      eris=approx_eris)
        approx_energy, _, _ = energy_function(approx_eris)
        thc_res_cost = compute_cost(
            num_spin_orbs,
            thc_lambda.lambda_total,
            thc_rank * num_spin_orbs // 2,
            list(kmesh),
            chi=chi,
            beta=beta,
            dE_for_qpe=dE_for_qpe,
        )
        thc_resource_obj.add_resources(
            ham_properties=thc_lambda,
            resource_estimates=thc_res_cost,
            cutoff=thc_rank,
            approx_energy=np.real(approx_energy),
        )

    return thc_resource_obj.to_dataframe()



#coverage:ignore
import os
from uuid import uuid4
import h5py
import numpy
import numpy.random
import numpy.linalg
from scipy.optimize import minimize
import jax
import jax.numpy as jnp
from jax.config import config
from jax import jit, grad
from .adagrad import adagrad
from .thc_objectives import (thc_objective, thc_objective_grad,
                             thc_objective_and_grad,
                             cp_ls_cholesky_factor_objective,
                             thc_objective_regularized)

# set mkl thread count for numpy einsum/tensordot calls
# leave one CPU un used  so we can still access this computer
os.environ["MKL_NUM_THREADS"] = "{}".format(os.cpu_count() - 1)
config.update("jax_enable_x64", True)


class CallBackStore:

    def __init__(self, chkpoint_file, freqency=500):
        """Generic callback function  for storing intermediates from BFGS and
           Adagrad optimizations
        """
        self.chkpoint_file = chkpoint_file
        self.freq = freqency
        self.iter = 0

    def __call__(self, xk):
        if self.iter % self.freq == 0:
            f = h5py.File(self.chkpoint_file, "w")
            f["xk"] = xk
            f.close()


def lbfgsb_opt_thc(eri,
                   nthc,
                   chkfile_name=None,
                   initial_guess=None,
                   random_seed=None,
                   maxiter=150_000,
                   disp=False):
    """
    Least-squares fit of two-electron integral tensors with  L-BFGS-B
    """
    # initialize chkfile name if one isn't set
    if chkfile_name is None:
        chkfile_name = str(uuid4()) + '.h5'

    # callback func stores checkpoints
    callback_func = CallBackStore(chkfile_name)

    # set initial guess
    norb = eri.shape[0]
    if initial_guess is None:
        if random_seed is None:
            x = numpy.random.randn(norb * nthc + nthc * nthc)
        else:
            numpy.random.seed(random_seed)
            x = numpy.random.randn(norb * nthc + nthc * nthc)
    else:
        x = initial_guess  # add more checks here for safety

    # L-BFGS-B optimization
    res = minimize(thc_objective_and_grad,
                   x,
                   args=(norb, nthc, eri),
                   jac=True,
                   method='L-BFGS-B',
                   options={
                       'disp': disp,
                       'maxiter': maxiter
                   },
                   callback=callback_func)
    # print(res)
    params = res.x
    x = numpy.array(params)
    f = h5py.File(chkfile_name, "w")
    f["etaPp"] = x[:norb * nthc].reshape(nthc, norb)
    f["ZPQ"] = x[norb * nthc:].reshape(nthc, nthc)
    f.close()
    return params


def lbfgsb_opt_thc_l2reg(eri,
                         nthc,
                         chkfile_name=None,
                         initial_guess=None,
                         random_seed=None,
                         maxiter=150_000,
                         disp_freq=98,
                         penalty_param=None,
                         disp=False):
    """
    Least-squares fit of two-electron integral tensors with  L-BFGS-B with
    l2-regularization of lambda

    disp is ignored.
    disp_freq sets the freqnecy of printing
    """
    if disp_freq > 98 or disp_freq < 1:
        raise ValueError(
            "disp_freq {} is not valid. must be between [1, 98]".format(
                disp_freq))

    if chkfile_name is None:
        #chkfile_name = str(uuid4()) + '.h5'
        callback_func = None
    else:
        # callback func stores checkpoints
        callback_func = CallBackStore(chkfile_name)

    # set initial guess
    norb = eri.shape[0]
    if initial_guess is None:
        if random_seed is None:
            x = numpy.random.randn(norb * nthc + nthc * nthc)
        else:
            numpy.random.seed(random_seed)
            x = numpy.random.randn(norb * nthc + nthc * nthc)
    else:
        x = initial_guess  # add more checks here for safety

    # compute inital lambda to set penalty param
    etaPp = x[:norb * nthc].reshape(nthc, norb)  # leaf tensor  nthc x norb
    MPQ = x[norb * nthc:norb * nthc + nthc * nthc].reshape(
        nthc, nthc)  # central tensor
    SPQ = etaPp.dot(
        etaPp.T)  # (nthc x norb)  x (norb x nthc) -> (nthc  x nthc) metric
    cP = jnp.diag(jnp.diag(
        SPQ))  # grab diagonal elements. equivalent to np.diag(np.diagonal(SPQ))
    # no sqrts because we have two normalized THC vectors (index by mu and nu)
    # on each side.
    MPQ_normalized = cP.dot(MPQ).dot(cP)  # get normalized zeta in Eq. 11 & 12
    lambda_z = jnp.sum(jnp.abs(MPQ_normalized)) * 0.5
    # lambda_z = jnp.sum(MPQ_normalized**2) * 0.5
    CprP = jnp.einsum("Pp,Pr->prP", etaPp,
                      etaPp)  # this is einsum('mp,mq->pqm', etaPp, etaPp)
    Iapprox = jnp.einsum('pqU,UV,rsV->pqrs',
                         CprP,
                         MPQ,
                         CprP,
                         optimize=[(0, 1), (0, 1)])
    deri = eri - Iapprox
    # set penalty
    if penalty_param is None:
        sum_square_loss = 0.5 * numpy.sum((deri)**2)
        penalty_param = sum_square_loss / lambda_z
        print("lambda_z {}".format(lambda_z))
        print("penalty_param {}".format(penalty_param))

    # L-BFGS-B optimization
    thc_grad = jax.grad(thc_objective_regularized, argnums=[0])
    print("Initial Grad")
    print(thc_grad(jnp.array(x), norb, nthc, jnp.array(eri), penalty_param))
    print()
    res = minimize(thc_objective_regularized,
                   jnp.array(x),
                   args=(norb, nthc, jnp.array(eri), penalty_param),
                   method='L-BFGS-B',
                   jac=thc_grad,
                   options={
                       'disp': None,
                       'iprint': disp_freq,
                       'maxiter': maxiter
                   },
                   callback=callback_func)

    # print(res)
    params = numpy.array(res.x)
    x = numpy.array(params)
    if chkfile_name is not None:
        f = h5py.File(chkfile_name, "w")
        f["etaPp"] = x[:norb * nthc].reshape(nthc, norb)
        f["ZPQ"] = x[norb * nthc:].reshape(nthc, nthc)
        f.close()
    return params


def adagrad_opt_thc(eri,
                    nthc,
                    chkfile_name=None,
                    initial_guess=None,
                    random_seed=None,
                    stepsize=0.01,
                    momentum=0.9,
                    maxiter=50_000,
                    gtol=1.0E-5):
    """
    THC opt usually starts with BFGS and then is completed with Adagrad or other
    first order solver.  This  function implements an Adagrad optimization.

    Optimization runs for 50 K iterations.  This is the ONLY stopping cirteria
    used in the FT-THC paper by Lee et al.
    """
    # initialize chkfile name if one isn't set
    if chkfile_name is None:
        chkfile_name = str(uuid4()) + '.h5'

    # callback func stores checkpoints
    callback_func = CallBackStore(chkfile_name)

    # set initial guess
    norb = eri.shape[0]
    if initial_guess is None:
        if random_seed is None:
            x = numpy.random.randn(norb * nthc + nthc * nthc)
        else:
            numpy.random.seed(random_seed)
            x = numpy.random.randn(norb * nthc + nthc * nthc)
    else:
        x = initial_guess  # add more checks here for safety
    opt_init, opt_update, get_params = adagrad(step_size=stepsize,
                                               momentum=momentum)
    opt_state = opt_init(x)

    def update(i, opt_state):
        params = get_params(opt_state)
        gradient = thc_objective_grad(params, norb, nthc, eri)
        grad_norm_l1 = numpy.linalg.norm(gradient, ord=1)
        return opt_update(i, gradient, opt_state), grad_norm_l1

    for t in range(maxiter):
        opt_state, grad_l1 = update(t, opt_state)
        params = get_params(opt_state)
        if t % callback_func.freq == 0:
            # callback_func(params)
            fval = thc_objective(params, norb, nthc, eri)
            outline = "Objective val {: 5.15f}".format(fval)
            outline += "\tGrad L1-norm {: 5.15f}".format(grad_l1)
            print(outline)
        if grad_l1 <= gtol:
            # break out of loop
            # which sends to save
            break
    else:
        print("Maximum number of iterations reached")
    # save results before returning
    x = numpy.array(params)
    f = h5py.File(chkfile_name, "w")
    f["etaPp"] = x[:norb * nthc].reshape(nthc, norb)
    f["ZPQ"] = x[norb * nthc:].reshape(nthc, nthc)
    f.close()
    return params


def lbfgsb_opt_cholesky(cholesky_factor,
                        nthc,
                        chkfile_name=None,
                        initial_guess=None,
                        random_seed=None):
    """
    Least-squares fit of cholesky tensors with  L-BFGS-B

    cholesky_factor is reshaped into (norb, norb, num_cholesky)
    """
    # initialize chkfile name if one isn't set
    if chkfile_name is None:
        chkfile_name = str(uuid4()) + '.h5'

    # callback func stores checkpoints
    callback_func = CallBackStore(chkfile_name)

    # set initial guess
    norb = cholesky_factor.shape[0]
    ncholfactor = cholesky_factor.shape[-1]

    if initial_guess is None:
        if random_seed is None:
            x = numpy.random.randn(norb * nthc + nthc * ncholfactor)
        else:
            numpy.random.seed(random_seed)
            x = numpy.random.randn(norb * nthc + nthc * ncholfactor)
    else:
        x = initial_guess  # add more checks here for safety

    # L-BFGS-B optimization
    res = minimize(cp_ls_cholesky_factor_objective,
                   x,
                   args=(norb, nthc, cholesky_factor, True),
                   jac=True,
                   method='L-BFGS-B',
                   options={
                       'disp': True,
                       'ftol': 1.0E-4
                   },
                   callback=callback_func)
    print(res)
    params = res.x
    x = numpy.array(params)
    f = h5py.File(chkfile_name, "w")
    f["etaPp"] = x[:norb * nthc].reshape(nthc, norb)
    f["gamma"] = x[norb * nthc:].reshape(nthc, cholesky_factor.shape[-1])
    f.close()
    return params



#coverage:ignore
"""Test cases for costing_sf.py
"""
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates import sf


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_reiher_sf():
    """ Reproduce Reiher et al orbital SF FT costs from paper """
    DE = 0.001
    CHI = 10

    # Reiher et al orbitals
    N = 108
    LAM = 4258.0
    L = 200

    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps.
    # Then we input that into the function again.
    output = sf.compute_cost(N, LAM, DE, L, CHI, stps=20000)
    stps1 = output[0]
    output = sf.compute_cost(N, LAM, DE, L, CHI, stps1)
    assert output == (14184, 94868988984, 3320)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_li_sf():
    """ Reproduce Li et al orbital SF FT costs from paper """
    DE = 0.001
    CHI = 10

    # Li et al orbitals
    N = 152
    LAM = 3071.8
    L = 275
    # Here we're using an initial calculation with a very rough estimate of the
    # number of steps to give a more accurate number of steps.
    # Then we input that into the function again.
    output = sf.compute_cost(N, LAM, DE, L, CHI, stps=20000)
    stps2 = output[0]
    output = sf.compute_cost(N, LAM, DE, L, CHI, stps2)
    assert output == (24396, 117714920508, 3628)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import itertools

import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import mp

    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)
    from openfermion.resource_estimates.pbc.df.df_integrals import (
        DFABKpointIntegrals,)
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_df_amat_bmat():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    nmo = mymp.nmo

    Luv = cholesky_from_df_ints(mymp)  # [kpt, kpt, naux, nao, nao]
    dfk_inst = DFABKpointIntegrals(Luv.copy(), mf)
    naux = dfk_inst.naux

    dfk_inst.double_factorize()

    nkpts = len(mf.kpts)
    for qidx, kidx in itertools.product(range(nkpts), repeat=2):
        Amats, Bmats = dfk_inst.build_A_B_n_q_k_from_chol(qidx, kidx)
        # check if Amats and Bmats have the correct size
        assert Amats.shape == (naux, 2 * nmo, 2 * nmo)
        assert Bmats.shape == (naux, 2 * nmo, 2 * nmo)

        # check if Amats and Bmats have the correct symmetry--Hermitian
        assert np.allclose(Amats, Amats.conj().transpose(0, 2, 1))
        assert np.allclose(Bmats, Bmats.conj().transpose(0, 2, 1))

        # check if we can recover the Cholesky vector from Amat
        k_minus_q_idx = dfk_inst.k_transfer_map[qidx, kidx]
        test_chol = dfk_inst.build_chol_part_from_A_B(kidx, qidx, Amats, Bmats)
        assert np.allclose(test_chol, dfk_inst.chol[kidx, k_minus_q_idx])

        # check if factorized is working numerically exact case
        assert np.allclose(dfk_inst.amat_n_mats[kidx, qidx], Amats)
        assert np.allclose(dfk_inst.bmat_n_mats[kidx, qidx], Bmats)

        for nn in range(Amats.shape[0]):
            w, v = np.linalg.eigh(Amats[nn, :, :])
            non_zero_idx = np.where(w > 1.0e-4)[0]
            w = w[non_zero_idx]
            v = v[:, non_zero_idx]
            assert len(w) <= 2 * nmo

    for qidx in range(nkpts):
        for nn in range(naux):
            for kidx in range(nkpts):
                eigs_a_fixed_n_q = dfk_inst.amat_lambda_vecs[kidx, qidx, nn]
                eigs_b_fixed_n_q = dfk_inst.bmat_lambda_vecs[kidx, qidx, nn]
                assert len(eigs_a_fixed_n_q) <= 2 * nmo
                assert len(eigs_b_fixed_n_q) <= 2 * nmo

    for kidx in range(nkpts):
        for kpidx in range(nkpts):
            for qidx in range(nkpts):
                kmq_idx = dfk_inst.k_transfer_map[qidx, kidx]
                kpmq_idx = dfk_inst.k_transfer_map[qidx, kpidx]
                exact_eri_block = dfk_inst.get_eri_exact(
                    [kidx, kmq_idx, kpmq_idx, kpidx])
                test_eri_block = dfk_inst.get_eri(
                    [kidx, kmq_idx, kpmq_idx, kpidx])
                assert np.allclose(exact_eri_block, test_eri_block)



""" Determine costs for sparse decomposition in QC
    Note this is WIP
"""
from typing import Tuple
import numpy as np
from numpy.lib.scimath import arccos, arcsin  # has analytc continuation to cplx
from openfermion.resource_estimates.utils import QI, power_two


def cost_sparse(n: int, lam: float, d: int, dE: float, chi: int,
                stps: int) -> Tuple[int, int, int]:
    """ Determine fault-tolerant costs using sparse decomposition in quantum
        chemistry

    Args:
        n (int) - the number of spin-orbitals
        lam (float) - the lambda-value for the Hamiltonian
        dE (float) - allowable error in phase estimation
        L (int) - the rank of the first decomposition
        Lxi (int) - the total number of eigenvectors
        chi (int) - equivalent to aleph_1 and aleph_2 in the document, the
            number of bits for the representation of the coefficients
        beta (int) - equivalent to beth in the document, the number of bits
            for the rotations
        stps (int) - an approximate number of steps to choose the precision
            of single qubit rotations in preparation of the equal superposition
            state

    Returns:
        step_cost (int) - Toffolis per step
        total_cost (int) - Total number of Toffolis
        ancilla_cost (int) - Total ancilla cost
    """

    # I think there is a bug in the mathematica notebook. It does not check if
    # 2 is a factor first, which it should, cf. the similar function in
    # costingdf.nb Below is correct using the power_two() function, to give
    # power of 2 that is a factor of d.
    eta = power_two(d)

    nN = np.ceil(np.log2(n // 2))

    m = chi + 8 * nN + 4  # Eq (A13)

    oh = [0] * 20

    nM = (np.ceil(np.log2(d)) - eta) / 2

    for p in range(2, 22):
        # JJG note: arccos arg may be > 1
        v = np.round(np.power(2,p+1) / (2 * np.pi) * arccos(np.power(2,nM) /\
            np.sqrt(d/2**eta)/2))
        oh[p-2] = np.real(stps * (1 / (np.sin(3 * arcsin(np.cos(v * 2 * np.pi /\
            np.power(2,p+1)) * \
            np.sqrt(d/2**eta) / np.power(2,nM)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br = int(np.argmin(oh) + 1) + 2

    # Hand selecting the k expansion factor
    k1 = 32

    # Equation (A17)
    cost = np.ceil(d/k1) + m * (k1 -1) + QI(d)[1] + 4 * n + 8 * nN + 2 * chi + \
        7 * np.ceil(np.log2(d)) - 6 * eta + 4 * br - 19

    # Number of iterations needed for the phase estimation.
    iters = np.ceil(np.pi * lam / (dE * 2))

    # The following are the number of qubits from the listing on page 39.

    # Control registers for phase estimation and iteration on them.
    ac1 = 2 * np.ceil(np.log2(iters)) - 1

    # System qubits
    ac2 = n

    # The register used for the QROM
    ac3 = np.ceil(np.log2(d))

    # The qubit used for flagging the success of the equal superposition state
    # preparation and the ancilla qubit for rotation
    ac45 = 2

    # The phase gradient state
    ac6 = br

    # The equal superposition state for coherent alias sampling
    ac7 = chi

    # The ancillas used for QROM
    ac8 = np.ceil(np.log2(d / k1)) + m * k1

    ancilla_cost = ac1 + ac2 + ac3 + ac45 + ac6 + ac7 + ac8

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert ancilla_cost.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(ancilla_cost)

    return step_cost, total_cost, ancilla_cost



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.resources.data_types import (
        PBCResources,
        ResourceEstimates,
    )
    from openfermion.resource_estimates.pbc.hamiltonian import (
        HamiltonianProperties,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_pbc_resources():
    nmo = 10
    np.random.seed(7)
    pbc_resources = PBCResources(
        "pbc",
        num_spin_orbitals=nmo,
        num_kpts=12,
        dE=1e-7,
        chi=13,
        exact_energy=-13.3,
    )
    for cutoff in np.logspace(-1, -3, 5):
        lv = np.random.random(3)
        lambdas = HamiltonianProperties(lambda_total=lv[0],
                                        lambda_one_body=lv[1],
                                        lambda_two_body=lv[2])
        resource = ResourceEstimates(
            toffolis_per_step=np.random.randint(0, 1000),
            total_toffolis=np.random.randint(0, 1000),
            logical_qubits=13,
        )
        pbc_resources.add_resources(
            ham_properties=lambdas,
            resource_estimates=resource,
            approx_energy=-12,
            cutoff=cutoff,
        )
    df = pbc_resources.to_dataframe()
    assert np.allclose(
        df.lambda_total.values,
        [0.07630829, 0.97798951, 0.26843898, 0.38094113, 0.21338535],
    )
    assert (df.toffolis_per_step.values == [919, 366, 895, 787, 949]).all()
    assert (df.total_toffolis.values == [615, 554, 391, 444, 112]).all()


#coverage:ignore
""" Single factorization of the ERI tensor """
import numpy as np
from openfermion.resource_estimates.utils import eigendecomp


def factorize(eri_full, rank):
    """ Do single factorization of the ERI tensor

    Args:
       eri_full (np.ndarray) - 4D (N x N x N x N) full ERI tensor
       rank (int) - number of vectors to retain in ERI rank-reduction procedure

    Returns:
       eri_rr (np.ndarray) - 4D approximate ERI tensor reconstructed from LR vec
       LR (np.ndarray) - 3D (N x N x rank) tensor containing SF vectors
    """
    n_orb = eri_full.shape[0]
    assert n_orb**4 == len(eri_full.flatten())

    L = eigendecomp(eri_full.reshape(n_orb**2, n_orb**2), tol=1e-16)

    # Do rank-reduction of ERIs following ERI factorization
    if rank is None:
        LR = L[:, :]
    else:
        LR = L[:, :rank]
    eri_rr = np.einsum('ik,kj->ij', LR, LR.T, optimize=True)
    eri_rr = eri_rr.reshape(n_orb, n_orb, n_orb, n_orb)
    LR = LR.reshape(n_orb, n_orb, -1)
    if rank is not None:
        try:
            assert LR.shape[2] == rank
        except AssertionError:
            sys.exit(
                "LR.shape:     %s\nrank: %s\nLR.shape and rank are inconsistent"
                % (LR.shape, rank))

    return eri_rr, LR



#coverage:ignore
#   Copyright 2020 Google LLC

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from .adagrad import (adagrad, constant, exponential_decay,
                          inverse_time_decay, make_schedule, piecewise_constant,
                          polynomial_decay)
    from .thc_factorization import (adagrad_opt_thc, lbfgsb_opt_cholesky,
                                    lbfgsb_opt_thc, lbfgsb_opt_thc_l2reg)
    from .thc_objectives import (cp_ls_cholesky_factor_objective, thc_objective,
                                 thc_objective_and_grad, thc_objective_grad,
                                 thc_objective_grad_jax, thc_objective_jax,
                                 thc_objective_regularized)


# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass, asdict
from typing import Tuple
import h5py
import numpy as np
import numpy.typing as npt

from pyscf import lib
from pyscf.ao2mo import _ao2mo
from pyscf.lib import logger
from pyscf.pbc.df import df
from pyscf.pbc.lib.kpts_helper import gamma_point
from pyscf.pbc.mp.kmp2 import _add_padding
from pyscf.pbc import mp, scf, gto


@dataclass
class HamiltonianProperties:
    """Lighweight descriptive data class to hold return values from
    compute_lambda functions.

    Attributes:
        lambda_total: Total lambda value (norm) of Hamiltonian.
        lambda_one_body: One-body lambda value (norm) of Hamiltonian.
        lambda_two_body: Two-body lambda value (norm) of Hamiltonian.
    """

    lambda_total: float
    lambda_one_body: float
    lambda_two_body: float

    dict = asdict


def build_hamiltonian(mf: "scf.KRHF") -> Tuple[npt.NDArray, npt.NDArray]:
    """Utility function to build one- and two-electron matrix elements from mean
    field object.

    Arguments:
        mf: pyscf KRHF object.

    Returns:
        hcore_mo: one-body Hamiltonian in MO basis.
        chol: 3-index RSGDF density fitted integrals.
    """
    # Build temporary mp2 object so MO coeffs can potentially be padded if mean
    # field solution yields different number of MOs per k-point.
    tmp_mp2 = mp.KMP2(mf)
    mo_coeff_padded = _add_padding(tmp_mp2, tmp_mp2.mo_coeff,
                                   tmp_mp2.mo_energy)[0]
    hcore_mo = np.asarray([
        C.conj().T @ hk @ C for (C, hk) in zip(mo_coeff_padded, mf.get_hcore())
    ])
    chol = cholesky_from_df_ints(tmp_mp2)
    return hcore_mo, chol


def cholesky_from_df_ints(mp2_inst, pad_mos_with_zeros=True) -> npt.NDArray:
    """Compute 3-center electron repulsion integrals, i.e. (L|ov),
    where `L` denotes DF auxiliary basis functions and `o` and `v` occupied and
    virtual canonical crystalline orbitals. Note that `o` and `v` contain kpt
    indices `ko` and `kv`, and the third kpt index `kL` is determined by
    the conservation of momentum.

    Note that if the number of mos differs at each k-point then this function
    will pad MOs with zeros to ensure contiguity.

    Args:
        mp2_inst: pyscf KMP2 instance.

    Returns:
        Lchol: 3-center DF ints, with shape (nkpts, nkpts, naux, nmo, nmo)
    """

    log = logger.Logger(mp2_inst.stdout, mp2_inst.verbose)

    if mp2_inst._scf.with_df._cderi is None:
        mp2_inst._scf.with_df.build()

    cell = mp2_inst._scf.cell
    if cell.dimension == 2:
        # 2D ERIs are not positive definite. The 3-index tensors are stored in
        # two part. One corresponds to the positive part and one corresponds
        # to the negative part. The negative part is not considered in the
        # DF-driven CCSD implementation.
        raise NotImplementedError

    # nvir = nmo - nocc
    nao = cell.nao_nr()

    mo_coeff = mp2_inst._scf.mo_coeff
    kpts = mp2_inst.kpts
    if pad_mos_with_zeros:
        mo_coeff = _add_padding(mp2_inst, mp2_inst.mo_coeff,
                                mp2_inst.mo_energy)[0]
        nmo = mp2_inst.nmo
    else:
        nmo = nao
        num_mo_per_kpt = np.array([C.shape[-1] for C in mo_coeff])
        if not (num_mo_per_kpt == nmo).all():
            log.info("Number of MOs differs at each k-point or is not the same "
                     "as the number of AOs.")
    nkpts = len(kpts)
    if gamma_point(kpts):
        dtype = np.double
    else:
        dtype = np.complex128
    dtype = np.result_type(dtype, *mo_coeff)
    Lchol = np.empty((nkpts, nkpts), dtype=object)

    cput0 = (logger.process_clock(), logger.perf_counter())

    bra_start = 0
    bra_end = nmo
    ket_start = nmo
    ket_end = 2 * nmo
    with h5py.File(mp2_inst._scf.with_df._cderi, "r") as f:
        kptij_lst = f["j3c-kptij"][:]
        tao = []
        ao_loc = None
        for ki, kpti in enumerate(kpts):
            for kj, kptj in enumerate(kpts):
                kpti_kptj = np.array((kpti, kptj))
                Lpq_ao = np.asarray(df._getitem(f, "j3c", kpti_kptj, kptij_lst))

                mo = np.hstack((mo_coeff[ki], mo_coeff[kj]))
                mo = np.asarray(mo, dtype=dtype, order="F")
                if dtype == np.double:
                    out = _ao2mo.nr_e2(
                        Lpq_ao,
                        mo,
                        (bra_start, bra_end, ket_start, ket_end),
                        aosym="s2",
                    )
                else:
                    # Note: Lpq.shape[0] != naux if linear dependency is found
                    # in auxbasis
                    if Lpq_ao[0].size != nao**2:  # aosym = 's2'
                        Lpq_ao = lib.unpack_tril(Lpq_ao).astype(np.complex128)
                    out = _ao2mo.r_e2(
                        Lpq_ao,
                        mo,
                        (bra_start, bra_end, ket_start, ket_end),
                        tao,
                        ao_loc,
                    )
                Lchol[ki, kj] = out.reshape(-1, nmo, nmo)

    log.timer_debug1("transforming DF-AO integrals to MO", *cput0)

    return Lchol


def build_momentum_transfer_mapping(cell: gto.Cell,
                                    kpoints: np.ndarray) -> np.ndarray:
    # Define mapping momentum_transfer_map[Q][k1] = k2 that satisfies
    # k1 - k2 + G = Q.
    a = cell.lattice_vectors() / (2 * np.pi)
    delta_k1_k2_Q = (kpoints[:, None, None, :] - kpoints[None, :, None, :] -
                     kpoints[None, None, :, :])
    delta_k1_k2_Q += kpoints[0][None, None, None, :]  # shift to center
    delta_dot_a = np.einsum("wx,kpQx->kpQw", a, delta_k1_k2_Q)
    int_delta_dot_a = np.rint(delta_dot_a)
    # Should be zero if transfer is statisfied (2*pi*n)
    mapping = np.where(
        np.sum(np.abs(delta_dot_a - int_delta_dot_a), axis=3) < 1e-10)
    num_kpoints = len(kpoints)
    momentum_transfer_map = np.zeros((num_kpoints,) * 2, dtype=np.int32)
    # Note index flip due to Q being first index in map but broadcasted last..
    momentum_transfer_map[mapping[1], mapping[0]] = mapping[2]

    return momentum_transfer_map



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import itertools
import numpy as np
import numpy.typing as npt

from pyscf.pbc import scf

from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)


# Single-Factorization
class SingleFactorization:

    def __init__(self,
                 cholesky_factor: npt.NDArray,
                 kmf: scf.HF,
                 naux: int = None):
        """Class defining single-factorized ERIs.

        Args:
            cholesky_factor: Cholesky factor tensor that is
                [nkpts, nkpts, naux, nao, nao].
                To see how to generate this see cholesky_from_df_ints
            kmf: pyscf k-object.  Currently only used to obtain the number of
                k-points.  Must have an attribute kpts which len(self.kmf.kpts)
                returns number of kpts.
        """
        self.chol = cholesky_factor
        self.kmf = kmf
        self.nk = len(self.kmf.kpts)
        if naux is None:
            naux = cholesky_factor[0, 0].shape[0]
        self.naux = naux
        self.nao = cholesky_factor[0, 0].shape[-1]
        k_transfer_map = build_momentum_transfer_mapping(
            self.kmf.cell, self.kmf.kpts)
        self.k_transfer_map = k_transfer_map

    def build_AB_from_chol(self, qidx: int):
        """Construct A and B matrices given Q-kpt idx.

        This constructs all matrices association with n-chol

        Args:
            qidx: index for momentum mode Q.

        Returns:
            A:  A matrix of size [naux, nmo * kpts, nmk * kpts]
            B:  B matrix of size [naux, nmo * kpts, nmk * kpts]
        """
        nmo = self.nao  # number of gaussians used
        naux = self.naux

        # now form A and B
        # First set up matrices to store. We will loop over Q index and
        # construct entire set of matrices index by n-aux.
        rho = np.zeros((naux, nmo * self.nk, nmo * self.nk),
                       dtype=np.complex128)

        for kidx in range(self.nk):
            k_minus_q_idx = self.k_transfer_map[qidx, kidx]
            for p, q in itertools.product(range(nmo), repeat=2):
                P = int(kidx * nmo + p)  # a^_{pK}
                Q = int(k_minus_q_idx * nmo + q)  # a_{q(K-Q)}
                rho[:, P, Q] += self.chol[
                    kidx, k_minus_q_idx][:naux, p,
                                         q]  # L_{pK, q(K-Q)}a^_{pK}a_{q(K-Q)}

        A = 0.5 * (rho + rho.transpose((0, 2, 1)).conj())
        B = 0.5j * (rho - rho.transpose((0, 2, 1)).conj())

        return A, B

    def get_eri(self, ikpts: list, check_eq=False):
        r"""Construct (pkp qkq| rkr sks) via cholesky factors.

        Note: 3-tensor L_{sks, rkr} = L_{rkr, sks}^{*}

        Args:
            ikpts: list of four integers representing the index of the kpoint
                in self.kmf.kpts
            check_eq: optional value to confirm a symmetry in the Cholesky
                vectors (Default value = False)

        Returns:
            eri: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = ikpts
        n = self.naux
        naux_pq = self.chol[ikp, ikq].shape[0]
        if n > naux_pq:
            print(f"WARNING: specified naux ({n}) is too large!")
            n = naux_pq
        if check_eq:
            assert np.allclose(
                np.einsum(
                    "npq,nsr->pqrs",
                    self.chol[ikp, ikq][:n],
                    self.chol[iks, ikr][:n].conj(),
                    optimize=True,
                ),
                np.einsum(
                    "npq,nrs->pqrs",
                    self.chol[ikp, ikq][:n],
                    self.chol[ikr, iks][:n],
                    optimize=True,
                ),
            )
        return np.einsum(
            "npq,nsr->pqrs",
            self.chol[ikp, ikq][:n],
            self.chol[iks, ikr][:n].conj(),
            optimize=True,
        )



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np

from openfermion.resource_estimates.pbc.resources.qrom import (QR2, QI2)


def test_qr2():
    L = 728
    npp = 182
    bpp = 21
    test_val = QR2(L + 1, npp, bpp)
    assert np.isclose(test_val, 3416)

    L = 56
    npp = 28
    bpp = 91
    test_val = QR2(L + 1, npp, bpp)
    assert np.isclose(test_val, 679)


def test_qi2():
    L1 = 728
    npp = 182
    test_val = QI2(L1 + 1, npp)
    assert np.isclose(test_val, 785)

    L1 = 56
    npp = 28
    test_val = QI2(L1 + 1, npp)
    assert np.isclose(test_val, 88)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Compute resource estimates for sparse LCU of k-point Hamiltonian
"""
from typing import Tuple
import numpy as np
from numpy.lib.scimath import arccos, arcsin  # has analytc continuation to cplx
from sympy import factorint
from openfermion.resource_estimates.utils import QI

from openfermion.resource_estimates.pbc.resources import (ResourceEstimates,
                                                          QR3)


def compute_cost(
        num_spin_orbs: int,
        lambda_tot: float,
        num_sym_unique: int,
        kmesh: list[int],
        dE_for_qpe: float = 0.0016,
        chi: int = 10,
) -> ResourceEstimates:
    """Determine fault-tolerant costs using sparse representaion of Hamiltonian.

    Light wrapper around _compute_cost to automate choice of stps paramter.

    Arguments:
        num_spin_orbs: the number of spin-orbitals
        lambda_tot: the lambda-value for the Hamiltonian
        num_sym_unique: number of symmetry unique terms kept in the sparse
            Hamiltonian
        dE_for_qpe: allowable error in phase estimation
        chi: the number of bits for the representation of the coefficients
    Returns:
        resources: sparse resources
    """
    # run once to determine stps parameter
    init_cost = _compute_cost(
        num_spin_orbs,
        lambda_tot,
        num_sym_unique,
        dE_for_qpe,
        chi,
        20_000,
        *kmesh,
    )
    steps = init_cost[0]
    final_cost = _compute_cost(
        num_spin_orbs,
        lambda_tot,
        num_sym_unique,
        dE_for_qpe,
        chi,
        steps,
        *kmesh,
    )
    estimates = ResourceEstimates(
        toffolis_per_step=final_cost[0],
        total_toffolis=final_cost[1],
        logical_qubits=final_cost[2],
    )
    return estimates


def _compute_cost(
        n: int,
        lam: float,
        d: int,
        dE: float,
        chi: int,
        stps: int,
        Nkx: int,
        Nky: int,
        Nkz: int,
) -> Tuple[int, int, int]:
    """Determine fault-tolerant costs using sparse representaion of Hamiltonian.

    Arguments:
        n: the number of spin-orbitals
        lam: the lambda-value for the Hamiltonian
        d: number of symmetry unique terms kept in the sparse Hamiltonian
        dE: allowable error in phase estimation
        chi: the number of bits for the representation of the coefficients
        stps: an approximate number of steps to choose the precision
            of single qubit rotations in preparation of the equal superposition
            state
        Nkx: number of k-points in x-direction
        Nky: number of k-points in y-direction
        Nkz: number of k-points in z-direction
    Returns:
        step_cost: Toffolis per step
        total_cost: Total number of Toffolis
        ancilla_cost: Total ancilla cost
    """
    if n % 2 != 0:
        raise ValueError("The number of spin orbitals is always even!")

    factors = factorint(d)
    eta = factors[min(list(sorted(factors.keys())))]
    if d % 2 == 1:
        eta = 0

    nN = np.ceil(np.log2(n // 2))
    nNk = (max(np.ceil(np.log2(Nkx)), 1) + max(np.ceil(np.log2(Nky)), 1) +
           max(np.ceil(np.log2(Nkz)), 1))
    Nk = Nkx * Nky * Nkz

    m = chi + 8 * nN + 6 * nNk + 5  # Eq 26

    oh = [0] * 20

    nM = (np.ceil(np.log2(d)) - eta) / 2

    for p in range(2, 22):
        # JJG note: arccos arg may be > 1
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nM) / np.sqrt(d / 2**eta) / 2))
        oh[p - 2] = np.real(stps * (1 / (np.sin(3 * arcsin(
            np.cos(v * 2 * np.pi / np.power(2, p + 1)) * np.sqrt(d / 2**eta) /
            np.power(2, nM)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br = int(np.argmin(oh) + 1) + 2

    # Hand selecting the k expansion factor
    k1 = QR3(d, m)[0]

    # Equation (A17)
    cost = (QR3(d, m)[1] + QI(d)[1] + 6 * n * Nk + 8 * nN + 12 * nNk + 2 * chi +
            7 * np.ceil(np.log2(d)) - 6 * eta + 4 * br - 8)

    # The following are adjustments if we don't need to do explicit arithmetic
    # to make subtraction modular
    if Nkx == 2**np.ceil(np.log2(Nkx)):
        cost = cost - 2 * np.ceil(np.log2(Nkx))
    if Nky == 2**np.ceil(np.log2(Nky)):
        cost = cost - 2 * np.ceil(np.log2(Nky))
    if Nkz == 2**np.ceil(np.log2(Nkz)):
        cost = cost - 2 * np.ceil(np.log2(Nkz))

    # Number of iterations needed for the phase estimation.
    iters = np.ceil(np.pi * lam / (dE * 2))

    # The following are the number of qubits from the listing on page 39.

    # Control registers for phase estimation and iteration on them.
    ac1 = 2 * np.ceil(np.log2(iters)) - 1

    # System qubits
    ac2 = n * Nk

    # The register used for the QROM
    ac3 = np.ceil(np.log2(d))

    # The phase gradient state
    ac6 = br

    # The equal superposition state for coherent alias sampling
    ac7 = chi

    # The ancillas used for QROM
    ac8 = np.ceil(np.log2(d / np.power(2, k1))) + m * np.power(2, k1)

    ac9 = 9

    ancilla_cost = ac1 + ac2 + ac3 + ac6 + ac7 + ac8 + ac9

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert ancilla_cost.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(ancilla_cost)

    return step_cost, total_cost, ancilla_cost



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.sparse.\
        compute_sparse_resources import _compute_cost, compute_cost


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_compute_cost():
    nRe = 108
    lam_re = 2135.3
    dRe = 705831
    dE = 0.001
    chi = 10

    nLi = 152
    lam_Li = 1547.3
    dLi = 440501
    dE = 0.001
    chi = 10

    Nkx = 2
    Nky = 2
    Nkz = 2

    res = _compute_cost(nRe, lam_re, dRe, dE, chi, 20_000, Nkx, Nky, Nkz)
    # res = _compute_cost(nRe, lam_re, dRe, dE, chi, res[0], Nkx, Nky, Nkz)
    assert res[0] == 22962
    assert res[1] == 77017349364
    assert res[2] == 11335

    res = _compute_cost(nRe, lam_re, dRe, dE, chi, 20_000, 3, 5, 1)
    assert res[0] == 29004
    assert res[1] == 97282954488
    assert res[2] == 8060

    res = _compute_cost(nLi, lam_Li, dLi, dE, chi, 20_000, Nkx, Nky, Nkz)
    res = _compute_cost(nLi, lam_Li, dLi, dE, chi, res[0], Nkx, Nky, Nkz)
    assert res[0] == 21426
    assert res[1] == 52075764444
    assert res[2] == 7015

    res = _compute_cost(nLi, lam_Li, dLi, dE, chi, 20_000, 3, 5, 1)
    assert res[0] == 28986
    assert res[1] == 70450299084
    assert res[2] == 9231


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_compute_cost_helper():
    dE = 0.001
    chi = 10

    nLi = 152
    lam_Li = 1547.3
    dLi = 440501
    dE = 0.001
    chi = 10

    Nkx = 2
    Nky = 2
    Nkz = 2

    res = compute_cost(
        num_spin_orbs=nLi,
        lambda_tot=lam_Li,
        num_sym_unique=dLi,
        kmesh=[Nkx, Nky, Nkz],
        dE_for_qpe=dE,
        chi=chi,
    )
    assert res.toffolis_per_step == 21426
    assert res.total_toffolis == 52075764444
    assert res.logical_qubits == 7015


if __name__ == "__main__":
    test_compute_cost()



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt

from openfermion.resource_estimates.pbc.hamiltonian import (
    HamiltonianProperties,)
from openfermion.resource_estimates.pbc.sparse.sparse_integrals import (
    SparseFactorization,)


@dataclass
class SparseHamiltonianProperties(HamiltonianProperties):
    """Store for return values of compute_lambda function

    Extension of HamiltonianProperties dataclass to also hold the number of
    retained matrix elements (num_sym_unique).
    """

    num_sym_unique: int


def compute_lambda(hcore: npt.NDArray, sparse_int_obj: SparseFactorization
                  ) -> SparseHamiltonianProperties:
    """Compute lambda value for sparse method

    Arguments:
        hcore: array of hcore(k) by kpoint. k-point order
            is pyscf order generated for this problem.
        sparse_int_obj: The sparse integral object that is used
            to compute eris and the number of unique
            terms.

    Returns:
        ham_props: A SparseHamiltonianProperties instance containing Lambda
            values for the sparse hamiltonian.
    """
    kpts = sparse_int_obj.kmf.kpts
    nkpts = len(kpts)
    one_body_mat = np.empty((len(kpts)), dtype=object)
    lambda_one_body = 0.0

    import time

    for kidx in range(len(kpts)):
        # matrices for - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
        # and  + 0.5 sum_{Q}sum_{r}(pkqk|rQrQ)
        h1_pos = np.zeros_like(hcore[kidx])
        h1_neg = np.zeros_like(hcore[kidx])
        for qidx in range(len(kpts)):
            # - 0.5 * sum_{Q}sum_{r}(pkrQ|rQqk)
            eri_kqqk_pqrs = sparse_int_obj.get_eri_exact(
                [kidx, qidx, qidx, kidx])
            h1_neg -= (np.einsum("prrq->pq", eri_kqqk_pqrs, optimize=True) /
                       nkpts)
            # + sum_{Q}sum_{r}(pkqk|rQrQ)
            eri_kkqq_pqrs = sparse_int_obj.get_eri_exact(
                [kidx, kidx, qidx, qidx])

            h1_pos += np.einsum("pqrr->pq", eri_kkqq_pqrs) / nkpts

        one_body_mat[kidx] = hcore[kidx] + 0.5 * h1_neg + h1_pos
        lambda_one_body += np.sum(np.abs(one_body_mat[kidx].real)) + np.sum(
            np.abs(one_body_mat[kidx].imag))

    lambda_two_body = 0.0
    nkpts = len(kpts)
    # recall (k, k-q|k'-q, k')
    for kidx in range(nkpts):
        for kpidx in range(nkpts):
            for qidx in range(nkpts):
                kmq_idx = sparse_int_obj.k_transfer_map[qidx, kidx]
                kpmq_idx = sparse_int_obj.k_transfer_map[qidx, kpidx]
                test_eri_block = (
                    sparse_int_obj.get_eri([kidx, kmq_idx, kpmq_idx, kpidx]) /
                    nkpts)
                lambda_two_body += np.sum(np.abs(test_eri_block.real)) + np.sum(
                    np.abs(test_eri_block.imag))

    lambda_tot = lambda_one_body + lambda_two_body
    sparse_data = SparseHamiltonianProperties(
        lambda_total=lambda_tot,
        lambda_one_body=lambda_one_body,
        lambda_two_body=lambda_two_body,
        num_sym_unique=sparse_int_obj.get_total_unique_terms_above_thresh(
            return_nk_counter=False),
    )
    return sparse_data



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for Richardson Gaudin model module."""

import pytest
import numpy as np

from openfermion.hamiltonians import RichardsonGaudin
from openfermion.ops import QubitOperator
from openfermion.transforms import get_fermion_operator
from openfermion.transforms import jordan_wigner
from openfermion.linalg import get_sparse_operator
from openfermion import(get_fermion_operator, InteractionOperator, \
                         normal_ordered)


@pytest.mark.parametrize('g, n_qubits, expected', [
    (0.3, 2,
     QubitOperator('3.0 [] + 0.15 [X0 X1] + \
0.15 [Y0 Y1] - 1.0 [Z0] - 2.0 [Z1]')),
    (-0.1, 3,
     QubitOperator('6.0 [] - 0.05 [X0 X1] - 0.05 [X0 X2] - \
0.05 [Y0 Y1] - 0.05 [Y0 Y2] - 1.0 [Z0] - 0.05 [X1 X2] - \
0.05 [Y1 Y2] - 2.0 [Z1] - 3.0 [Z2]')),
])
def test_richardson_gaudin_hamiltonian(g, n_qubits, expected):
    rg = RichardsonGaudin(g, n_qubits)
    rg_qubit = rg.qubit_operator
    assert rg_qubit == expected

    assert np.array_equal(
        np.sort(np.unique(get_sparse_operator(rg_qubit).diagonal())),
        2 * np.array(list(range((n_qubits + 1) * n_qubits // 2 + 1))))


def test_n_body_tensor_errors():
    rg = RichardsonGaudin(1.7, n_qubits=2)
    with pytest.raises(TypeError):
        rg.n_body_tensors = 0
    with pytest.raises(TypeError):
        rg.constant = 1.1


@pytest.mark.parametrize("g, n_qubits", [(0.2, 4), (-0.2, 4)])
def test_fermionic_hamiltonian_from_integrals(g, n_qubits):
    rg = RichardsonGaudin(g, n_qubits)
    #hc, hr1, hr2 = rg.hc, rg.hr1, rg.hr2
    doci = rg
    constant = doci.constant
    reference_constant = 0

    doci_qubit_op = doci.qubit_operator
    doci_mat = get_sparse_operator(doci_qubit_op).toarray()
    doci_eigvals = np.linalg.eigh(doci_mat)[0]

    tensors = doci.n_body_tensors
    one_body_tensors, two_body_tensors = tensors[(1, 0)], tensors[(1, 1, 0, 0)]

    fermion_op = get_fermion_operator(
        InteractionOperator(constant, one_body_tensors, 0.5 * two_body_tensors))
    fermion_op = normal_ordered(fermion_op)
    fermion_mat = get_sparse_operator(fermion_op).toarray()
    fermion_eigvals = np.linalg.eigh(fermion_mat)[0]

    one_body_tensors2, two_body_tensors2 = rg.get_antisymmetrized_tensors()
    fermion_op2 = get_fermion_operator(
        InteractionOperator(reference_constant, one_body_tensors2,
                            0.5 * two_body_tensors2))
    fermion_op2 = normal_ordered(fermion_op2)
    fermion_mat2 = get_sparse_operator(fermion_op2).toarray()
    fermion_eigvals2 = np.linalg.eigh(fermion_mat2)[0]

    for eigval in doci_eigvals:
        assert any(abs(fermion_eigvals -
                       eigval) < 1e-6), "The DOCI spectrum should have \
        been contained in the spectrum of the fermionic operator constructed via the \
        DOCIHamiltonian class"

    for eigval in doci_eigvals:
        assert any(abs(fermion_eigvals2 -
                       eigval) < 1e-6), "The DOCI spectrum should have \
       been contained in the spectrum of the fermionic operators constructed via the anti \
       symmetrized tensors"



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np

from openfermion.resource_estimates.pbc.thc.factorizations.kmeans import (
    KMeansCVT)


def gaussian(dx, sigma):
    return np.exp(-np.einsum("Gx,Gx->G", dx, dx) /
                  sigma**2.0) / (2 * (sigma * np.pi)**0.5)


def gen_gaussian(xx, yy, grid, sigma=0.125):
    x0 = np.array([0.25, 0.25])
    dx0 = grid - x0[None, :]
    weight = gaussian(dx0, sigma)
    x1 = np.array([0.25, 0.75])
    dx = grid - x1[None, :]
    weight += gaussian(dx, sigma)
    x2 = np.array([0.75, 0.25])
    dx = grid - x2[None, :]
    weight += gaussian(dx, sigma)
    x3 = np.array([0.75, 0.75])
    dx = grid - x3[None, :]
    weight += gaussian(dx, sigma)
    return weight


def test_kmeans():
    np.random.seed(7)
    # 3D spatial grid
    num_grid_x = 10
    num_grid_points = num_grid_x**2
    xs = np.linspace(0, 1, num_grid_x)
    ys = np.linspace(0, 1, num_grid_x)
    xx, yy = np.meshgrid(xs, ys)
    grid = np.zeros((num_grid_points, 2))
    grid[:, 0] = xx.ravel()
    grid[:, 1] = yy.ravel()
    num_interp_points = 10

    weight = gen_gaussian(xx, yy, grid)
    kmeans = KMeansCVT(grid)
    interp_points = kmeans.find_interpolating_points(num_interp_points,
                                                     weight,
                                                     verbose=False)
    interp_points_ref = [37, 27, 77, 81, 38, 24, 73, 62, 76, 22]
    assert np.allclose(interp_points, interp_points_ref)


if __name__ == "__main__":
    test_kmeans()



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os

from pyscf.pbc import gto, scf
from pyscf.lib import chkfile


def make_diamond_113_szv() -> scf.KRHF:
    """Build diamond in gth-szv basis with 113 k-point mesh from checkpoint
    file.
    """
    TEST_CHK = os.path.join(os.path.dirname(__file__), "../test_data/scf.chk")
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.build(parse_arg=False)

    kmesh = [1, 1, 3]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts).rs_density_fit()
    scf_dict = chkfile.load(TEST_CHK, "scf")
    mf.__dict__.update(scf_dict)
    mf.with_df._cderi = TEST_CHK
    mf.with_df.mesh = cell.mesh
    return mf



#coverage:ignore
#   Copyright 2020 Google LLC

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from .compute_cost_df import compute_cost
    from .compute_lambda_df import compute_lambda
    from .factorize_df import factorize
    from .generate_costing_table_df import generate_costing_table


# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import mp

    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.testing.systems import (
        make_diamond_113_szv,)
    from openfermion.resource_estimates.pbc.sparse.sparse_integrals import (
        unique_iter, unique_iter_pr_qs, unique_iter_ps_qr, unique_iter_pq_rs,
        SparseFactorization)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_sparse_int_obj():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    Luv = cholesky_from_df_ints(mymp)
    for thresh in [1.0e-3, 1.0e-4, 1.0e-5, 1.0e-6]:
        abs_sum_coeffs = 0
        helper = SparseFactorization(cholesky_factor=Luv,
                                     kmf=mf,
                                     threshold=thresh)
        nkpts = len(mf.kpts)
        # recall (k, k-q|k'-q, k')
        for kidx in range(nkpts):
            for kpidx in range(nkpts):
                for qidx in range(nkpts):
                    kmq_idx = helper.k_transfer_map[qidx, kidx]
                    kpmq_idx = helper.k_transfer_map[qidx, kpidx]
                    test_eri_block = helper.get_eri(
                        [kidx, kmq_idx, kpmq_idx, kpidx])
                    abs_sum_coeffs += np.sum(np.abs(
                        test_eri_block.real)) + np.sum(
                            np.abs(test_eri_block.imag))


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_get_num_unique():
    mf = make_diamond_113_szv()
    mymp = mp.KMP2(mf)
    Luv = cholesky_from_df_ints(mymp)
    helper = SparseFactorization(cholesky_factor=Luv, kmf=mf)

    import itertools
    from pyscf.pbc.lib.kpts_helper import KptsHelper, loop_kkk

    # import the iteration routines
    nmo = mf.mo_coeff[0].shape[-1]
    tally4 = np.zeros((nmo, nmo, nmo, nmo), dtype=int)
    for ft in unique_iter(nmo):
        p, q, r, s = ft
        if p == q == r == s:
            tally4[p, q, r, s] += 1
        elif p == r and q == s:
            tally4[p, q, r, s] += 1
            tally4[q, p, s, r] += 1
        elif p == s and q == r:
            tally4[p, q, r, s] += 1
            tally4[q, p, s, r] += 1
        elif p == q and r == s:
            tally4[p, q, r, s] += 1
            tally4[r, s, p, q] += 1
        else:
            tally4[p, q, r, s] += 1
            tally4[q, p, s, r] += 1
            tally4[s, r, q, p] += 1
            tally4[r, s, p, q] += 1
    assert np.allclose(tally4, 1)

    kpts_helper = KptsHelper(mf.cell, mf.kpts)
    nkpts = len(mf.kpts)
    completed = np.zeros((nkpts, nkpts, nkpts), dtype=bool)
    tally = np.zeros((nkpts, nkpts, nkpts), dtype=int)
    fulltally = np.zeros((nkpts, nkpts, nkpts, nmo, nmo, nmo, nmo), dtype=int)
    nk_count = 0
    for kvals in loop_kkk(nkpts):
        kp, kq, kr = kvals
        ks = kpts_helper.kconserv[kp, kq, kr]
        if not completed[kp, kq, kr]:
            eri_block = helper.get_eri([kp, kq, kr, ks])
            nk_count += 1
            if kp == kq == kr == ks:
                completed[kp, kq, kr] = True
                tally[kp, kq, kr] += 1
                for ftuple in unique_iter(
                        nmo
                ):  # iterate over unique whenever all momentum indices are the
                    # same
                    p, q, r, s = ftuple
                    if p == q == r == s:
                        fulltally[kp, kq, kr, p, q, r, s] += 1
                    elif p == r and q == s:
                        fulltally[kp, kq, kr, p, q, r, s] += 1
                        fulltally[kp, kq, kr, q, p, s, r] += 1
                    elif p == s and q == r:
                        fulltally[kp, kq, kr, p, q, r, s] += 1
                        fulltally[kp, kq, kr, q, p, s, r] += 1
                    elif p == q and r == s:
                        fulltally[kp, kq, kr, p, q, r, s] += 1
                        fulltally[kp, kq, kr, r, s, p, q] += 1
                    else:
                        fulltally[kp, kq, kr, p, q, r, s] += 1
                        fulltally[kp, kq, kr, q, p, s, r] += 1
                        fulltally[kp, kq, kr, s, r, q, p] += 1
                        fulltally[kp, kq, kr, r, s, p, q] += 1

            elif kp == kq and kr == ks:
                completed[kp, kq, kr] = True
                completed[kr, ks, kp] = True
                tally[kp, kq, kr] += 1
                tally[kr, ks, kp] += 1

                # the full (kp kq | kr ks) gets mapped to (kr, ks | kp  kq)
                # fulltally[kp, kq, kr] += 1
                # fulltally[kr, ks, kp] += 1

                # we only need to count the single eri block because
                # (kp, kq kr ks) -> (kr, ks, kp kq) 1:1
                # but for (p q | r s) -> (qp|sr) so we are overcounting
                # by a little (q p | s r)
                # since we have a complex conjugation symmetry in both we really
                # have (npair, npair) here. we can use pyscf.ao2mo to do this.

                test_block = np.zeros_like(eri_block, dtype=int)
                num_terms_in_block = 0
                for p, q, r, s in unique_iter_pq_rs(nmo):
                    num_terms_in_block += 1
                    test_block[p, q, r, s] += 1
                    fulltally[kp, kq, kr, p, q, r, s] += 1
                    fulltally[kr, ks, kp, r, s, p, q] += 1
                    if p == q and r == s:
                        continue
                    else:
                        test_block[q, p, s, r] += 1
                        fulltally[kp, kq, kr, q, p, s, r] += 1
                        fulltally[kr, ks, kp, s, r, q, p] += 1

                for p, q, r, s in itertools.product(range(helper.nao),
                                                    repeat=4):
                    if not np.isclose(test_block[p, q, r, s], 1):
                        print(p, q, r, s, test_block[p, q, r, s])
                assert np.allclose(test_block, 1)

                assert num_terms_in_block <= helper.nao**4

            elif kp == ks and kq == kr:
                completed[kp, kq, kr] = True
                completed[kr, ks, kp] = True
                tally[kp, kq, kr] += 1
                tally[kr, ks, kp] += 1

                # fulltally[kp, kq, kr] += 1
                # fulltally[kr, ks, kp] += 1

                test_block = np.zeros_like(eri_block, dtype=int)
                num_terms_in_block = 0
                for p, q, r, s in unique_iter_ps_qr(nmo):
                    num_terms_in_block += 1
                    test_block[p, q, r, s] += 1
                    fulltally[kp, kq, kr, p, q, r, s] += 1
                    fulltally[kr, ks, kp, r, s, p, q] += 1
                    if p == s and q == r:
                        continue
                    else:
                        test_block[s, r, q, p] += 1
                        fulltally[kp, kq, kr, s, r, q, p] += 1
                        fulltally[kr, ks, kp, q, p, s, r] += 1

                for p, q, r, s in itertools.product(range(helper.nao),
                                                    repeat=4):
                    if not np.isclose(test_block[p, q, r, s], 1):
                        print(p, q, r, s, test_block[p, q, r, s])
                assert np.allclose(test_block, 1)

                assert num_terms_in_block <= helper.nao**4

            elif kp == kr and kq == ks:
                completed[kp, kq, kr] = True
                completed[kq, kp, ks] = True
                tally[kp, kq, kr] += 1
                tally[kq, kp, ks] += 1
                # symmetry takes account of [kq, kp, ks] only need to do one of
                # the blocks

                test_block = np.zeros_like(eri_block, dtype=int)
                num_terms_in_block = 0
                for p, q, r, s in unique_iter_pr_qs(nmo):
                    num_terms_in_block += 1
                    test_block[p, q, r, s] += 1
                    fulltally[kp, kq, kr, p, q, r, s] += 1
                    fulltally[kq, kp, ks, q, p, s, r] += 1
                    if p == r and q == s:
                        continue
                    else:
                        test_block[r, s, p, q] += 1
                        fulltally[kp, kq, kr, r, s, p, q] += 1
                        fulltally[kq, kp, ks, s, r, q, p] += 1

                for p, q, r, s in itertools.product(range(helper.nao),
                                                    repeat=4):
                    if not np.isclose(test_block[p, q, r, s], 1):
                        print(p, q, r, s, test_block[p, q, r, s])
                assert np.allclose(test_block, 1)

                assert num_terms_in_block <= helper.nao**4

            else:
                completed[kp, kq, kr] = True
                completed[kr, ks, kp] = True
                completed[kq, kp, ks] = True
                completed[ks, kr, kq] = True

                tally[kp, kq, kr] += 1
                tally[kr, ks, kp] += 1
                tally[kq, kp, ks] += 1
                tally[ks, kr, kq] += 1

                # just assign entire 4-tensor +1 value because each pqrs is
                # unique because kp, kq, kr, ks is unique for this case we
                # would only need to grab one of these blocks of 4.
                fulltally[kp, kq, kr] += 1
                fulltally[kq, kp, ks] += 1
                fulltally[ks, kr, kq] += 1
                fulltally[kr, ks, kp] += 1

    assert np.allclose(completed, True)
    assert np.allclose(tally, 1)

    for kvals in loop_kkk(nkpts):
        kp, kq, kr = kvals
        ks = kpts_helper.kconserv[kp, kq, kr]
        if len(set([kp, kq, kr, ks])) == 4:
            # print(kp, kq, kr, np.allclose(fulltally[kp, kq, kr], 1))
            assert np.allclose(fulltally[kp, kq, kr], 1)
        elif kp == kr and kq == ks:
            assert np.allclose(fulltally[kp, kq, kr], 1)
            assert np.allclose(fulltally[kq, kp, ks], 1)
        elif kp == ks and kq == kr:
            assert np.allclose(fulltally[kp, kq, kr], 1)
            assert np.allclose(fulltally[kr, ks, kp], 1)
        elif kp == kq and kr == ks:
            assert np.allclose(fulltally[kp, kq, kr], 1)
            assert np.allclose(fulltally[kr, ks, kp], 1)

        assert np.allclose(fulltally[kp, kp, kp], 1)
